from bilby.gw.conversion import convert_to_lal_binary_black_hole_parameters
from bilby.gw.conversion import _generate_all_cbc_parameters


def dQuadMonS_and_dQuadMonA_to_dQuadMon1_and_dQuadMon2(dQuadMonS, dQuadMonA):
    """
    Convert symmetric and anti-symmetric combinations of spin-induced quadrupole
    moment parameters to that of binary components.

    Parameters
    ----------
    dQuadMonS: float
            symmetric combination of individual spin-induced quadrupole moment params, dQuadMon1 and dQuadMon2
    dQuadMonA: float
            anti-symmetric combination of individual spin-induced quadrupole moment params, dQuadMon1 and dQuadMon2
    Return
    ------
    dQuadMon1: float
        Spin-induced quadrupole parameter of the heavier object
    dQuadMon2: float
        Spin-induced quadrupole parameter of the lighter object

    """

    dQuadMon1 = dQuadMonS + dQuadMonA
    dQuadMon2 = dQuadMonS - dQuadMonA
    return dQuadMon1, dQuadMon2


def dQuadMon1_and_dQuadMon2_to_dQuadMonS_and_dQuadMonA(dQuadMon1, dQuadMon2):
    """
    Convert spin-induced quadrupole moment params to their symmetric and
    anti symmetric combinations.

    Parameters
    ----------
    dQuadMon1: float
        Spin-induced quadrupole parameter of the heavier object
    dQuadMon2: float
        Spin-induced quadrupole parameter of the lighter object

    Return
    ------
    dQuadMonS: float
        symmetric combination of individual spin-induced quadrupole moment params, dQuadMon1 and dQuadMon2
    dQuadMonA: float
        anti-symmetric combination of individual spin-induced quadrupole moment params, dQuadMon1 and dQuadMon2
    """

    dQuadMonS = 0.5 * (dQuadMon1 + dQuadMon2)
    dQuadMonA = 0.5 * (dQuadMon1 - dQuadMon2)
    return dQuadMonS, dQuadMonA


def convert_to_lal_siqm_parameters(parameters):
    """
    Convert parameters we have into parameters we need.

    This is defined by the parameters of bilby.source.lal_binary_black_hole()


    Mass: mass_1, mass_2
    Spin: a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl
    Extrinsic: luminosity_distance, theta_jn, phase, ra, dec, geocent_time, psi

    This involves popping a lot of things from parameters.
    The keys in added_keys should be popped after evaluating the waveform.

    Parameters
    ----------
    parameters: dict
        dictionary of parameter values to convert into the required parameters

    Return
    ------
    converted_parameters: dict
        dict of the required parameters
    added_keys: list
        keys which are added to parameters during function call
    """

    converted_parameters, added_keys = convert_to_lal_binary_black_hole_parameters(
        parameters
    )

    if not any(
        [
            key in converted_parameters
            for key in ["dQuadMon1", "dQuadMon2", "dQuadMonS", "dQuadMonA"]
        ]
    ):
        converted_parameters["dQuadMon1"] = 0
        converted_parameters["dQuadMon2"] = 0
        added_keys = added_keys + ["dQuadMon1", "dQuadMon2"]
        return converted_parameters, added_keys

    if (
        "dQuadMonS" in converted_parameters.keys()
        and "dQuadMonA" in converted_parameters.keys()
    ):
        (
            converted_parameters["dQuadMon1"],
            converted_parameters["dQuadMon2"],
        ) = dQuadMonS_and_dQuadMonA_to_dQuadMon1_and_dQuadMon2(
            converted_parameters["dQuadMonS"], parameters["dQuadMonA"]
        )

    return converted_parameters, added_keys


def generate_all_siqm_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    From either a single sample or a set of samples fill in all missing
    BBH parameters, in place.

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    """
    waveform_defaults = {
        "reference_frequency": 50.0,
        "waveform_approximant": "IMRPhenomPv2",
        "minimum_frequency": 20.0,
    }
    output_sample = _generate_all_cbc_parameters(
        sample,
        defaults=waveform_defaults,
        base_conversion=convert_to_lal_siqm_parameters,
        likelihood=likelihood,
        priors=priors,
        npool=npool,
    )
    return output_sample



import unittest
import bilby_tgr


class TestBasicConversions(unittest.TestCase):
    def setUp(self):
        self.dQuadMon1 = 10.0
        self.dQuadMon2 = 10.0
        self.dQuadMonS = 0.5 * (self.dQuadMon1 + self.dQuadMon2)
        self.dQuadMonA = 0.5 * (self.dQuadMon1 - self.dQuadMon2)

    def test_dQuadMon1_and_dQuadMon2_to_dQuadMonS_and_dQuadMonA(self):
        (
            dQuadMonS,
            dQuadMonA,
        ) = bilby_tgr.siqm.conversion.dQuadMon1_and_dQuadMon2_to_dQuadMonS_and_dQuadMonA(
            self.dQuadMon1, self.dQuadMon2
        )
        self.assertAlmostEqual(self.dQuadMonS, dQuadMonS)
        self.assertAlmostEqual(self.dQuadMonA, dQuadMonA)

    def test_dQuadMonS_and_dQuadMonA_to_dQuadMon1_and_dQuadMon2(self):
        (
            dQuadMon1,
            dQuadMon2,
        ) = bilby_tgr.siqm.conversion.dQuadMonS_and_dQuadMonA_to_dQuadMon1_and_dQuadMon2(
            self.dQuadMonS, self.dQuadMonA
        )
        self.assertAlmostEqual(self.dQuadMon1, dQuadMon1)
        self.assertAlmostEqual(self.dQuadMon2, dQuadMon2)


class TestConvertToLALParams(unittest.TestCase):
    def setUp(self):
        self.search_keys = []
        self.parameters = dict()
        self.component_siqm_parameters = dict(dQuadMon1=10.0, dQuadMon2=10.0)
        self.all_component_pars = self.component_siqm_parameters.copy()
        self.siqm_parameters = self.component_siqm_parameters.copy()
        self.siqm_parameters[
            "dQuadMonS", "dQuadMonA"
        ] = bilby_tgr.siqm.conversion.dQuadMon1_and_dQuadMon2_to_dQuadMonS_and_dQuadMonA(
            **self.all_component_pars
        )

    def tearDown(self):
        del self.search_keys
        del self.parameters

    def siqm_convert(self):
        (
            self.parameters,
            self.added_keys,
        ) = bilby_tgr.siqm.conversion.convert_to_lal_siqm_parameters(self.parameters)

    def _conversion_to_component_siqm(self, keys):
        for key in keys:
            self.parameters[key] = self.siqm_parameters[key]
        self.siqm_convert()
        component_dict = {
            key: self.parameters[key] for key in ["dQuadMon1", "dQuadMon2"]
        }
        self.assertDictEqual(component_dict, self.component_siqm_parameters)

    def test_dQuadMon1_and_dQuadMon2_to_dQuadMonS_and_dQuadMonA(self):
        self._conversion_to_component_siqm(["dQuadMon1", "dQuadMon2"])


class TestGenerateAllParameters(unittest.TestCase):
    def setUp(self):
        self.parameters = dict(
            mass_1=36.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.5,
            tilt_2=1.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=2000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
            dQuadMonS=10,
            dQuadMonA=0,
        )
        self.expected_bbh_keys = [
            "mass_1",
            "mass_2",
            "a_1",
            "a_2",
            "tilt_1",
            "tilt_2",
            "phi_12",
            "phi_jl",
            "luminosity_distance",
            "theta_jn",
            "psi",
            "phase",
            "geocent_time",
            "ra",
            "dec",
            "reference_frequency",
            "waveform_approximant",
            "minimum_frequency",
            "chirp_mass",
            "total_mass",
            "symmetric_mass_ratio",
            "mass_ratio",
            "iota",
            "spin_1x",
            "spin_1y",
            "spin_1z",
            "spin_2x",
            "spin_2y",
            "spin_2z",
            "phi_1",
            "phi_2",
            "chi_eff",
            "chi_1_in_plane",
            "chi_2_in_plane",
            "chi_p",
            "cos_tilt_1",
            "cos_tilt_2",
            "redshift",
            "comoving_distance",
            "mass_1_source",
            "mass_2_source",
            "chirp_mass_source",
            "total_mass_source",
        ]

        self.expected_siqm_keys = [
            "dQuadMon1",
            "dQuadMon2",
            "dQuadMonS",
            "dQuadMonA",
        ]

    def test_generate_all_siqm_parameters(self):
        new_parameters = bilby_tgr.siqm.conversion.generate_all_siqm_parameters(
            self.parameters
        )
        for key in self.expected_bbh_keys + self.expected_siqm_keys:
            self.assertIn(key, new_parameters)


if __name__ == "__main__":
    unittest.main()



from .conversion import generate_liv_parameters, redshift_to_LIV_distance
from bilby.gw.conversion import luminosity_distance_to_redshift
import numpy as np


def get_reweighting_weights(posterior, prior_initial, prior_final='default', cosmology=None):
    """
    From Jacobians of transformations, calculates weight for reweighting

    Parameters
    ==========
    posterior: pandas.DataFrame
        Bilby LIV posterior to transform
    prior_initial: string, one of A_eff, log_A_eff, log_lambda_eff
        String specifying prior in which LIV parameter was sampled
    prior_final: string, one of uniform, mass_uniform, default
        String specifying desired prior on A_alpha for reweighting
    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.


    Output

    weighted: pandas.DataFrame
        relative weights of the samples

    """

    assert ((prior_initial in ['A_eff', 'log_A_eff', 'log_lambda_eff', 'mass_graviton_eff', 'mdr_phase']) and
            (prior_final in ['uniform', 'mass_uniform', 'default'])), "Unknown prior"
    posterior_reweighted = posterior.copy()
    if prior_final == 'default':
        if prior_initial == 'mass_graviton_eff':
            prior_final = 'mass_uniform'
        else:
            prior_final = 'uniform'
    if 'A_alpha' not in posterior_reweighted:
        posterior_reweighted = generate_liv_parameters(posterior_reweighted)
    if 'A_alpha' not in posterior_reweighted:
        print('ERROR: Could not find LIV parameter in the posterior')

    # get distance measures for reweighting
    if 'redshift' in posterior_reweighted:
        z = posterior_reweighted['redshift']
    else:
        z = luminosity_distance_to_redshift(
            posterior_reweighted['luminosity_distance'], cosmology=cosmology)
    LIV_distance = redshift_to_LIV_distance(z, posterior_reweighted['alpha'], cosmology=cosmology)

    # distance ratio often useful in reweighting formulas
    distance_ratio = (posterior_reweighted['luminosity_distance'] / LIV_distance *
                      (1 + z)**(1 - posterior_reweighted['alpha']))

    # weights for reweighting from jacobian of the transformation (up to multiplicative constant)
    if prior_final == 'uniform':
        if prior_initial == 'A_eff':
            weights = distance_ratio
        elif prior_initial == 'mass_graviton_eff':
            weights = np.abs(distance_ratio * posterior_reweighted['A_alpha'])**0.5
            # discard negative mass samples (shouldn't be there anyway)
            weights[posterior_reweighted['A_alpha'] < 0] = 0
        elif prior_initial == 'mdr_phase':
            weights = distance_ratio / posterior_reweighted['luminosity_distance']
        else:
            weights = np.abs(posterior_reweighted['A_alpha'])
    elif prior_final == 'mass_uniform':
        assert np.all(posterior_reweighted['alpha'] == 0.0), 'mass_uniform prior is compatible only with alpha=0'
        if prior_initial == 'mass_graviton_eff':
            weights = distance_ratio**0.5
        elif prior_initial == 'A_eff':
            weights = distance_ratio**0.5 / np.abs(posterior_reweighted['A_eff'])**0.5
            # discard negative mass samples (shouldn't be there anyway)
            weights[posterior_reweighted['A_alpha'] < 0] = 0
        else:
            weights = np.abs(posterior_reweighted['A_alpha'])**0.5
    return weights


def reweighted_pdf_1D(posterior, prior_initial, prior_final='uniform',
                      parameter='A_alpha', n_bins=10000, cosmology=None):
    """
    Reweight 1D LIV posterior between prior on effective LIV parameter and prior on the physical
    LIV parameter, and smooth it out with kde.
    Use only with parameters that do NOT have sharp posterior cutoff / periodic bounds

    Parameters
    ==========
    posterior: pandas.DataFrame
        Bilby LIV posterior to transform
    prior_initial: string, one of A_eff, log_A_eff, log_lambda_eff
        String specifying prior in which LIV parameter was sampled
    prior_final: string, one of uniform, mass_uniform,
        String specifying desired prior on A_alpha for reweighting
    parameter: string,
        String specifying the name of the parameter to transform
    n_bins: int,
        Number of interpolation points

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.


    Output

    interpolator: interpolation function for computing pdf of the given parameter
    p_range: tuple giving the minimum and maximum value of the parameter among the posterior samples

    """
    from scipy.stats import gaussian_kde
    from scipy.interpolate import interp1d

    # Check if the posterior is compatible with LIV analysis
    assert ((prior_initial in ['A_eff', 'log_A_eff', 'log_lambda_eff', 'mass_graviton_eff']) and
            (prior_final in ['uniform', 'mass_uniform', 'default'])), "Unknown prior"
    weights = get_reweighting_weights(posterior, prior_initial, prior_final)

    # make pdf using kde
    samples = posterior[parameter].values.copy()
    pdf = gaussian_kde(samples, weights=weights)
    n_bins = 10000
    p_min = min(samples)
    p_max = max(samples)
    x = np.linspace(2 * min(p_min, 0), 2 * p_max, n_bins)

    return interp1d(x, pdf(x), fill_value=0., bounds_error=False), (p_min, p_max)



import bilby
import numpy as np
from bilby.gw.cosmology import get_cosmology
from scipy import constants
from bilby.core.utils.constants import speed_of_light as c, parsec as pc
from scipy.constants import physical_constants

h = physical_constants["Planck constant in eV/Hz"][0] * 1e12

# helper functions from astropy


def isiterable(obj):
    """
    Returns True if the given object is iterable.
    Copied from astropy, as location varies between versions
    """

    try:
        iter(obj)
        return True
    except TypeError:
        return False


def vectorize_if_needed(func, *x):
    """
    Helper function to vectorize functions on array inputs
    Copied here from astropy, because vectorization differs between astropy versions
    """
    if any(map(isiterable, x)):
        return np.vectorize(func)(*x)
    else:
        return func(*x)

# distance conversions - converts between distances and LIV distance measure
# https://arxiv.org/pdf/1110.2720.pdf (eq. 15) written to be of the same form as bilby.gw.conversion functions


def luminosity_distance_to_redshift_scalar(distance, cosmology=None):
    """
    copied from bilby, but non-vectorized
    """
    from astropy import units
    from astropy.cosmology import z_at_value
    cosmology = get_cosmology(cosmology)
    return z_at_value(cosmology.luminosity_distance, distance * units.Mpc).value


def redshift_to_LIV_distance_scalar(z, alpha, cosmology=None):
    """
    adapting from astropy.cosmology integral_comoving_distance
    returns LIV distance measure https://arxiv.org/pdf/1110.2720.pdf (eq. 15)

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use DEFAULT_COSMOLOGY from bilby
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.
    """
    from scipy.integrate import quad
    cosmology = get_cosmology(cosmology)

    def integrand(x, *args):
        return cosmology._inv_efunc_scalar(x, *args) * (1 + x)**(alpha - 2.0)

    return (quad(integrand, 0.0, z, args=cosmology._inv_efunc_scalar_args)[0] *
            cosmology.hubble_distance.value * (1 + z)**(1 - alpha))


def redshift_to_LIV_distance(z, alpha, cosmology=None):
    """
    vector version of conversion to LIV distance

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use DEFAULT_COSMOLOGY from bilby
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology class.
    """
    if isiterable(z):
        z = np.asarray(z)
    if isiterable(alpha):
        alpha = np.asarray(alpha)
    return vectorize_if_needed(redshift_to_LIV_distance_scalar, z, alpha, cosmology)


def luminosity_distance_to_LIV_distance_scalar(distance, alpha, cosmology=None):
    z = luminosity_distance_to_redshift_scalar(distance, cosmology)
    return redshift_to_LIV_distance_scalar(z, alpha, cosmology)


def luminosity_distance_to_LIV_distance(distance, alpha, cosmology=None):
    z = bilby.gw.conversion.luminosity_distance_to_redshift(distance, cosmology)
    return redshift_to_LIV_distance(z, alpha, cosmology)


def comoving_distance_to_LIV_distance(distance, alpha, cosmology=None):
    z = bilby.gw.conversion.comoving_distance_to_redshift(distance, cosmology)
    return redshift_to_LIV_distance(z, alpha, cosmology)


# conversions between LIV parametrizations


def A_to_lambda(A_alpha, alpha, **kwargs):
    """
    Converts lorentz violation parameter expressed in terms of lambda_eff to one expressed in terms of A_alpha

    Parameters:

    A_alpha: np.array or float
        A_alpha in units of peV^(2-alpha)
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use DEFAULT_COSMOLOGY from bilby
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    lambda_true: np.array or float
        lambda in units of m
    """

    if isiterable(A_alpha):
        A_alpha = np.asarray(A_alpha)
    if isiterable(alpha):
        alpha = np.asarray(alpha)

    if ((isiterable(alpha) and (2. in alpha)) or (not isiterable(alpha) and alpha == 2.)):
        raise Exception("Error, exponent alpha has value 2, which is degenerate case")
    else:
        # hc in peV*m
        hc = constants.physical_constants["Planck constant in eV/Hz"][0] * constants.c * 1e12
        lambda_true = hc * A_alpha**(1. / (alpha - 2))
        return lambda_true


def lambda_to_A(lambda_true, alpha, **kwargs):
    """
    Parameters:

    lambda_true: np.array or float
        lambda in units of m
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use DEFAULT_COSMOLOGY from bilby
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    A_alpha: np.array or float
        A_alpha in units of peV^(2-alpha)
    """

    if isiterable(lambda_true):
        lambda_true = np.asarray(lambda_true)
    if isiterable(alpha):
        alpha = np.asarray(alpha)

    if ((isiterable(alpha) and (2. in alpha)) or (not isiterable(alpha) and alpha == 2.)):
        raise Exception("Error, exponent alpha has value 2, which is degenerate case")
    else:
        # hc in peV*m
        hc = constants.physical_constants["Planck constant in eV/Hz"][0] * constants.c * 1e12
        A_alpha = (lambda_true / hc)**(alpha - 2)
        return A_alpha


def lambda_to_lambda_eff(lambda_true, luminosity_distance, alpha, cosmology=None, **kwargs):
    """
    Parameters:

    lambda_true: np.array or float
        lambda in units of m
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use DEFAULT bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    lambda_eff: np.array or float
        lambda_eff in units of m
    """

    if isiterable(lambda_true):
        lambda_true = np.asarray(lambda_true)
    if isiterable(luminosity_distance):
        luminosity_distance = np.asarray(luminosity_distance)
    if isiterable(alpha):
        alpha = np.asarray(alpha)

    if ((isiterable(alpha) and (2. in alpha)) or (not isiterable(alpha) and alpha == 2.)):
        raise Exception("Error, exponent alpha has value 2, which is degenerate case")
    else:
        z = bilby.gw.conversion.luminosity_distance_to_redshift(luminosity_distance, cosmology)
        D_alpha = redshift_to_LIV_distance(z, alpha, cosmology)
        lambda_eff = ((1 + z)**(1 - alpha) * luminosity_distance / D_alpha)**(1. / (2 - alpha)) * lambda_true
        return lambda_eff


def lambda_eff_to_lambda(lambda_eff, luminosity_distance, alpha, redshift=None, cosmology=None, **kwargs):
    """
    Parameters:

    lambda_eff: np.array or float
        lambda_eff in units of m
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    lambda_true: np.array or float
        lambda in units of m
    """

    if isiterable(lambda_eff):
        lambda_eff = np.asarray(lambda_eff)
    if isiterable(luminosity_distance):
        luminosity_distance = np.asarray(luminosity_distance)
    if isiterable(alpha):
        alpha = np.asarray(alpha)

    if ((isiterable(alpha) and (2. in alpha)) or (not isiterable(alpha) and alpha == 2.)):
        raise Exception("Error, exponent alpha has value 2, which is degenerate case")
    else:
        if redshift is not None:
            z = redshift
        else:
            z = bilby.gw.conversion.luminosity_distance_to_redshift(luminosity_distance, cosmology)
        D_alpha = redshift_to_LIV_distance(z, alpha, cosmology)
        lambda_true = ((1 + z)**(1 - alpha) * luminosity_distance / D_alpha)**(1. / (alpha - 2)) * lambda_eff
        return lambda_true


def A_to_lambda_eff(A_alpha, luminosity_distance, alpha, cosmology=None, **kwargs):
    """
    Parameters:

    A_alpha: np.array or float
        A_alpha in peV^(2-alpha)
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    lambda_eff: np.array or float
        lambda_eff in units of m
    """

    lambda_eff = lambda_to_lambda_eff(A_to_lambda(A_alpha, alpha), luminosity_distance, alpha, cosmology)
    return lambda_eff


def lambda_eff_to_A(lambda_eff, luminosity_distance, alpha, redshift=None, cosmology=None, **kwargs):
    """
    Parameters:

    lambda_eff: np.array or float
        lambda_eff in units of m
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    A_alpha: np.array or float
        A_alpha in peV^(2-alpha)
    """

    A_alpha = lambda_to_A(
        lambda_eff_to_lambda(lambda_eff, luminosity_distance, alpha, redshift=redshift, cosmology=cosmology), alpha)
    return A_alpha


def log_lambda_eff_to_A(log_lambda_eff, luminosity_distance, alpha, sign_A, redshift=None, cosmology=None, **kwargs):
    return sign_A * lambda_eff_to_A(
        10.**log_lambda_eff, luminosity_distance, alpha, redshift=redshift, cosmology=cosmology, **kwargs)


# effective A


def A_to_A_eff(A_alpha, luminosity_distance, alpha, cosmology=None, **kwargs):
    """
    Parameters:

    A_alpha: np.array or float
        A_alpha in peV^(2-alpha)
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    A_eff: np.array or float
        A_eff in units of peV^(2-alpha)
    """
    z = bilby.gw.conversion.luminosity_distance_to_redshift(luminosity_distance, cosmology)
    LIV_distance = redshift_to_LIV_distance(z, alpha, cosmology=None)
    return LIV_distance * (1. + z)**(alpha - 1.) / luminosity_distance * A_alpha


def A_eff_to_A(A_eff, luminosity_distance, alpha, redshift=None, cosmology=None, **kwargs):
    """
    Parameters:

    A_eff: np.array or float
        A_eff in peV^(2-alpha)
    luminosity_distance: np.array or float
        luminosity distance in Mpc
    alpha: np.array or float
        exponent of lorentz violating momentum term

    redshift: np.array
        if given, skips redshift calculation step

    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
            None - Use default bilby cosmology
            Instance of astropy.cosmology.FLRW subclass
            String with name of known Astropy cosmology, e.g., Planck13
            Dictionary with arguments required to instantiate the cosmology
            class.

    Output

    A_alpha: np.array or float
        A_alpha in units of peV^(2-alpha)
    """
    if redshift is not None:
        z = redshift
    else:
        z = bilby.gw.conversion.luminosity_distance_to_redshift(luminosity_distance, cosmology)
    LIV_distance = redshift_to_LIV_distance(z, alpha, cosmology=None)
    return luminosity_distance * (1. + z)**(1. - alpha) / LIV_distance * A_eff


def mass_graviton_to_A(mass_graviton, **kwargs):
    """
    Parameters:

    mass_graviton: np.array or float
        mass of the graviton in peV

    Output

    A: np.array or float
        A in units of peV^(2-alpha)
    """
    return mass_graviton**2


def A_to_mass_graviton(A_alpha, **kwargs):
    """
    Parameters:

    A: np.array or float
        A in units of peV^(2-alpha)

    Output

    mass_graviton: np.array or float
        mass of the graviton in peV
    """
    return A_alpha**0.5


def mdr_phase_to_A_eff(mdr_phase, luminosity_distance, **kwargs):
    """
    Case alpha = 1 corresponds to constant phase. Conversion for this parametrization
    Parameters:

    mdr_phase: np.array or float
        Constant phase offset introduced by dispersion
    luminosity_distance: np.array or float
        luminosity distance in Mpc

    Output

    A_eff: np.array or float
        A_eff in units of peV^(2-alpha)
    """
    distance = luminosity_distance * pc * 1e6
    A_eff = - mdr_phase * c * h / distance / np.pi
    return A_eff


def A_eff_to_mdr_phase(A_eff, luminosity_distance, **kwargs):
    """
    Case alpha = 1 corresponds to constant phase. Conversion for this parametrization
    Parameters:

    A_eff: np.array or float
        A_eff in units of peV^(2-alpha)
    luminosity_distance: np.array or float
        luminosity distance in Mpc

    Output:

    mdr_phase: np.array or float
        Constant phase offset introduced by dispersion
    """
    distance = luminosity_distance * pc * 1e6
    mdr_phase = -(np.pi * distance * A_eff / h / c)
    return mdr_phase


def generate_all_bbh_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    adapted from bilby
    From either a single sample or a set of samples fill in all missing
    BBH parameters, in place.

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    """
    output_sample = bilby.gw.conversion.generate_all_bbh_parameters(sample, likelihood, priors, npool)
    output_sample = generate_liv_parameters(output_sample)
    return output_sample


def generate_all_bns_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    copied from bilby
    From either a single sample or a set of samples fill in all missing
    BNS parameters, in place.

    Since we assume BNS waveforms are aligned, component spins won't be
    calculated.

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    npool: int, (default=1)
        If given, perform generation (where possible) using a multiprocessing pool
    """
    output_sample = bilby.gw.conversion.generate_all_bns_parameters(sample, likelihood, priors, npool)
    output_sample = generate_liv_parameters(output_sample)
    return output_sample


def generate_liv_parameters(sample, cosmology=None):
    output_sample = sample.copy()
    redshift = None
    if 'redshift' in output_sample:
        redshift = output_sample['redshift']
    if 'alpha' in output_sample:
        if 'mass_graviton_eff' in output_sample:
            output_sample['A_eff'] = mass_graviton_to_A(output_sample["mass_graviton_eff"])
        elif 'mdr_phase' in output_sample:
            output_sample['A_eff'] = mdr_phase_to_A_eff(
                output_sample["mdr_phase"], output_sample['luminosity_distance'])
        if 'A_eff' in output_sample:
            output_sample['A_alpha'] = A_eff_to_A(
                A_eff=output_sample['A_eff'], luminosity_distance=output_sample['luminosity_distance'],
                alpha=output_sample['alpha'], redshift=redshift, cosmology=cosmology)
        elif 'log_lambda_eff' in output_sample and 'sign_A' in output_sample:
            output_sample['A_alpha'] = log_lambda_eff_to_A(
                log_lambda_eff=output_sample['log_lambda_eff'],
                luminosity_distance=output_sample['luminosity_distance'],
                alpha=output_sample['alpha'], sign_A=output_sample['sign_A'],
                redshift=redshift, cosmology=cosmology)

        if np.all(output_sample['alpha'] == 0.0):  # checks if alpha = 0, so if we consider graviton case
            if 'mass_graviton' not in output_sample:
                output_sample['mass_graviton'] = A_to_mass_graviton(output_sample["A_alpha"])
        if np.all(output_sample['alpha'] == 1.0):  # checks if alpha = 1, so constant phase case
            if 'mdr_phase' not in output_sample and 'A_eff' in output_sample:
                output_sample['mdr_phase'] = A_eff_to_mdr_phase(
                    output_sample["A_eff"], output_sample['luminosity_distance'])
    if 'A_alpha' not in output_sample:
        print('Could not generate A_alpha parameter')
    return output_sample


def convert_to_lal_binary_black_hole_parameters(parameters):
    """
    Convert parameters we have into parameters we need.

    This is defined by the parameters of bilby.source.lal_binary_black_hole()


    Mass: mass_1, mass_2
    Spin: a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl
    Extrinsic: luminosity_distance, theta_jn, phase, ra, dec, geocent_time, psi
    MDR: A_eff, alpha

    This involves popping a lot of things from parameters.
    The keys in added_keys should be popped after evaluating the waveform.

    Parameters
    ==========
    parameters: dict
        dictionary of parameter values to convert into the required parameters

    Returns
    =======
    converted_parameters: dict
        dict of the required parameters
    added_keys: list
        keys which are added to parameters during function call
    """

    converted_parameters, added_keys = bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters(parameters)
    if 'A_eff' not in converted_parameters:
        if 'mass_graviton_eff' in converted_parameters:
            converted_parameters['A_eff'] = mass_graviton_to_A(converted_parameters["mass_graviton_eff"])
            added_keys.append('A_eff')
        elif 'log_lambda_eff' in converted_parameters and 'sign_A' in converted_parameters:
            converted_parameters['A_eff'] = converted_parameters['sign_A'] * lambda_to_A(
                10**converted_parameters['log_lambda_eff'], converted_parameters['alpha'])
            added_keys.append('A_eff')
        elif 'mdr_phase' in converted_parameters:
            converted_parameters['A_eff'] = mdr_phase_to_A_eff(converted_parameters["mdr_phase"],
                                                               converted_parameters["luminosity_distance"])
            added_keys.append('A_eff')
    return converted_parameters, added_keys


def convert_to_lal_binary_neutron_star_parameters(parameters):
    """
    Convert parameters we have into parameters we need.

    This is defined by the parameters of bilby.source.lal_binary_black_hole()


    Mass: mass_1, mass_2
    Spin: a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl
    Extrinsic: luminosity_distance, theta_jn, phase, ra, dec, geocent_time, psi
    Tidal: lambda_1, lamda_2, lambda_tilde, delta_lambda_tilde
    MDR: A_eff, alpha

    This involves popping a lot of things from parameters.
    The keys in added_keys should be popped after evaluating the waveform.
    For details on tidal parameters see https://arxiv.org/pdf/1402.5156.pdf.

    Parameters
    ==========
    parameters: dict
        dictionary of parameter values to convert into the required parameters

    Returns
    =======
    converted_parameters: dict
        dict of the required parameters
    added_keys: list
        keys which are added to parameters during function call
    """

    converted_parameters, added_keys = bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters(parameters)
    if 'A_eff' not in converted_parameters:
        if 'mass_graviton_eff' in converted_parameters:
            converted_parameters['A_eff'] = mass_graviton_to_A(converted_parameters["mass_graviton_eff"])
            added_keys.append('A_eff')
        elif 'log_lambda_eff' in converted_parameters and 'sign_A' in converted_parameters:
            converted_parameters['A_eff'] = converted_parameters['sign_A'] * lambda_to_A(
                10**converted_parameters['log_lambda_eff'], converted_parameters['alpha'])
            added_keys.append('A_eff')
        elif 'mdr_phase' in converted_parameters:
            converted_parameters['A_eff'] = mdr_phase_to_A_eff(converted_parameters["mdr_phase"],
                                                               converted_parameters["luminosity_distance"])
            added_keys.append('A_eff')
    return converted_parameters, added_keys



import lalsimulation
from bilby.gw.source import _base_lal_cbc_fd_waveform
from lal import CreateDict


def lal_binary_black_hole(
    frequency_array,
    mass_1,
    mass_2,
    luminosity_distance,
    a_1,
    tilt_1,
    phi_12,
    a_2,
    tilt_2,
    phi_jl,
    theta_jn,
    phase,
    dQuadMon1,
    dQuadMon2,
    **kwargs,
):
    """Introducing spin-induced quadrupole moment terms (dQuadMon1 and dQuadMon2)
        to the lal_binary_black_hole source  model using lalsimulation

    Parameters
    ----------
    frequency_array: array_like
        The frequencies at which we want to calculate the strain
    mass_1: float
        The mass of the heavier object in solar masses
    mass_2: float
        The mass of the lighter object in solar masses
    luminosity_distance: float
        The luminosity distance in megaparsec
    a_1: float
        Dimensionless primary spin magnitude
    tilt_1: float
        Primary tilt angle
    phi_12: float
        Azimuthal angle between the two component spins
    a_2: float
        Dimensionless secondary spin magnitude
    tilt_2: float
        Secondary tilt angle
    phi_jl: float
        Azimuthal angle between the total binary angular momentum and the
        orbital angular momentum
    theta_jn: float
        Angle between the total binary angular momentum and the line of sight
    phase: float
        The phase at coalescence
    dQuadMon1: float
        The spin-induced quadrupole moment parameter of heavier object
    dQuadMon2: float
        The spin-induced quadrupole moment parameter of lighter object
    kwargs: dict
        Optional keyword arguments
        Supported arguments:
            waveform_approximant
            reference_frequency
            minimum_frequency
            maximum_frequency
            catch_waveform_errors
            pn_spin_order
            pn_tidal_order
            pn_phase_order
            pn_amplitude_order
            mode_array:
                Activate a specific mode array and evaluate the model using those
                modes only.  e.g., waveform_arguments =
                dict(waveform_approximant='IMRPhenomHM', mode_array=[[2,2],[2,-2])
                returns the 22 and 2-2 modes only of IMRPhenomHM.  You can only
                specify modes that are included in that particular model.  e.g.,
                waveform_arguments = dict(waveform_approximant='IMRPhenomHM',
                mode_array=[[2,2],[2,-2],[5,5],[5,-5]]) is not allowed because the
                55 modes are not included in this model.  Be aware that some models
                only take positive modes and return the positive and the negative
                mode together, while others need to call both.  e.g.,
                waveform_arguments = dict(waveform_approximant='IMRPhenomHM',
                mode_array=[[2,2],[4,-4]]) returns the 22 and 2-2 of IMRPhenomHM.
                However, waveform_arguments =
                dict(waveform_approximant='IMRPhenomXHM', mode_array=[[2,2],[4,-4]])
                returns the 22 and 4-4 of IMRPhenomXHM.

    Returns
    -------
    dict: A dictionary with the plus and cross polarisation strain modes
    """
    waveform_kwargs = dict(
        waveform_approximant="IMRPhenomPv2",
        reference_frequency=50.0,
        minimum_frequency=20.0,
        maximum_frequency=frequency_array[-1],
        catch_waveform_errors=False,
        pn_spin_order=-1,
        pn_tidal_order=-1,
        pn_phase_order=-1,
        pn_amplitude_order=0,
    )
    waveform_kwargs.update(kwargs)
    wf_dict = waveform_kwargs.get("lal_waveform_dictionary", CreateDict())
    lalsimulation.SimInspiralWaveformParamsInsertdQuadMon1(wf_dict, float(dQuadMon1))
    lalsimulation.SimInspiralWaveformParamsInsertdQuadMon2(wf_dict, float(dQuadMon2))

    waveform_kwargs["lal_waveform_dictionary"] = wf_dict

    return _base_lal_cbc_fd_waveform(
        frequency_array,
        mass_1=mass_1,
        mass_2=mass_2,
        luminosity_distance=luminosity_distance,
        a_1=a_1,
        tilt_1=tilt_1,
        phi_12=phi_12,
        a_2=a_2,
        tilt_2=tilt_2,
        phi_jl=phi_jl,
        theta_jn=theta_jn,
        phase=phase,
        dQuadMon1=dQuadMon1,
        dQuadMon2=dQuadMon2,
        **waveform_kwargs,
    )



import bilby.gw.prior as prior
import os
from bilby.core.utils import logger


DEFAULT_PRIOR_DIR = os.path.join(os.path.dirname(__file__), 'prior_files')


class BBHPriorDict(prior.CBCPriorDict):
    def __init__(self, dictionary=None, filename=None, aligned_spin=True,
                 conversion_function=None):
        """
        (Adapted from bilby)
        Initialises a Prior set for Binary Black holes with FTI parameters.

        Parameters
        ==========
        dictionary: dict, optional
            See superclass
        filename: str, optional
            See superclass
        conversion_function: func
            Function to convert between sampled parameters and constraints.
            By default this generates many additional parameters, see
            BBHPriorDict.default_conversion_function
        """
        if dictionary is None and filename is None:
            if aligned_spin:
                fname = 'aligned_spins_bbh.prior'
                logger.info('Using aligned spin prior')
            else:
                fname = 'precessing_spins_bbh.prior'
            filename = os.path.join(DEFAULT_PRIOR_DIR, fname)
            logger.info('No prior given, using default BBH priors in {}.'.format(filename))
        elif filename is not None:
            if not os.path.isfile(filename):
                filename = os.path.join(DEFAULT_PRIOR_DIR, filename)
        super(BBHPriorDict, self).__init__(dictionary=dictionary, filename=filename,
                                           conversion_function=conversion_function)

    def test_redundancy(self, key, disable_logging=False):
        """
        (Adapted from bilby)
        Test whether adding the key would be redundant.
        Already existing keys return True.

        Parameters
        ==========
        key: str
            The key to test.
        disable_logging: bool, optional
            Disable logging in this function call. Default is False.

        Returns
        =======
        redundant: bool
            Whether the key is redundant or not
        """
        if key in self:
            logger.debug('{} already in prior'.format(key))
            return True

        sampling_parameters = {key for key in self if not isinstance(
            self[key], prior.Constraint)}

        mass_parameters = {'mass_1', 'mass_2', 'chirp_mass', 'total_mass', 'mass_ratio', 'symmetric_mass_ratio'}
        spin_tilt_1_parameters = {'tilt_1', 'cos_tilt_1'}
        spin_tilt_2_parameters = {'tilt_2', 'cos_tilt_2'}
        spin_azimuth_parameters = {'phi_1', 'phi_2', 'phi_12', 'phi_jl'}
        inclination_parameters = {'theta_jn', 'cos_theta_jn'}
        distance_parameters = {'luminosity_distance', 'comoving_distance', 'redshift'}
        siqm_parameters = {'dchi_kappaS', 'dchi_kappaA', 'dchi_kappa1', 'dchi_kappa2'}

        if key in {'dchi_kappaS', 'dchi_kappaA'}:
            if 'dchi_kappa1' in sampling_parameters or 'dchi_kappa2' in sampling_parameters:
                logger.disabled = False
                logger.warning("Incompatible dkappas provided. This may lead to unexpected behaviour. \
                                Provide (dchi_kappaS and dchi_kappaA) xor (dchi_kappa1 and dchi_kappa2) instead.")
                return True
        if key in {'dchi_kappa1', 'dchi_kappa2'}:
            if 'dchi_kappaS' in sampling_parameters or 'dchi_kappaA' in sampling_parameters:
                logger.disabled = False
                logger.warning("Incompatible dkappas provided. This may lead to unexpected behaviour. \
                                Provide (dchi_kappaS and dchi_kappaA) xor (dchi_kappa1 and dchi_kappa2) instead.")
                return True

        for independent_parameters, parameter_set in \
                zip([2, 2, 1, 1, 1, 1, 2],
                    [mass_parameters, spin_azimuth_parameters,
                     spin_tilt_1_parameters, spin_tilt_2_parameters,
                     inclination_parameters, distance_parameters,
                     siqm_parameters]):
            if key in parameter_set:
                if len(parameter_set.intersection(
                        sampling_parameters)) >= independent_parameters:
                    logger.disabled = disable_logging
                    logger.warning('{} already in prior. '
                                   'This may lead to unexpected behaviour.'
                                   .format(parameter_set.intersection(self)))
                    logger.disabled = False
                    return True

        return False


class BNSPriorDict(prior.CBCPriorDict):
    def __init__(self, dictionary=None, filename=None, aligned_spin=True,
                 conversion_function=None):
        """
        (Adapted from bilby)
        Initialises a Prior set for Binary Neutron Stars with FTI parameters.

        Parameters
        ==========
        dictionary: dict, optional
            See superclass
        filename: str, optional
            See superclass
        conversion_function: func
            Function to convert between sampled parameters and constraints.
            By default this generates many additional parameters, see
            BNSPriorDict.default_conversion_function
        """
        if aligned_spin:
            default_file = 'aligned_spins_bns_tides_on.prior'
        else:
            default_file = 'precessing_spins_bns_tides_on.prior'
        if dictionary is None and filename is None:
            filename = os.path.join(DEFAULT_PRIOR_DIR, default_file)
            logger.info('No prior given, using default BNS priors in {}.'.format(filename))
        elif filename is not None:
            if not os.path.isfile(filename):
                filename = os.path.join(DEFAULT_PRIOR_DIR, filename)
        super(BNSPriorDict, self).__init__(dictionary=dictionary, filename=filename,
                                           conversion_function=conversion_function)

    def test_redundancy(self, key, disable_logging=False):
        logger.disabled = disable_logging
        logger.info("Performing redundancy check using BBHPriorDict(self).test_redundancy")
        bbh_redundancy = BBHPriorDict(self).test_redundancy(key)
        logger.disabled = False

        if bbh_redundancy:
            return True
        redundant = False

        sampling_parameters = {key for key in self if not isinstance(
            self[key], (prior.Constraint))}

        tidal_parameters = \
            {'lambda_1', 'lambda_2', 'lambda_tilde', 'delta_lambda_tilde'}

        if key in tidal_parameters:
            if len(tidal_parameters.intersection(sampling_parameters)) > 2:
                redundant = True
                logger.disabled = disable_logging
                logger.warning('{} already in prior. '
                               'This may lead to unexpected behaviour.'
                               .format(tidal_parameters.intersection(self)))
                logger.disabled = False
            elif len(tidal_parameters.intersection(sampling_parameters)) == 2:
                redundant = True
        return redundant

    @property
    def tidal(self):
        """ Return true if priors include phase parameters """
        return self.is_nonempty_intersection("tidal")


# (Adapted from bilby)
prior.Prior._default_latex_labels = {
    'mass_1': '$m_1$',
    'mass_2': '$m_2$',
    'total_mass': '$M$',
    'chirp_mass': r'$\mathcal{M}$',
    'mass_ratio': '$q$',
    'symmetric_mass_ratio': r'$\eta$',
    'a_1': '$a_1$',
    'a_2': '$a_2$',
    'tilt_1': r'$\theta_1$',
    'tilt_2': r'$\theta_2$',
    'cos_tilt_1': r'$\cos\theta_1$',
    'cos_tilt_2': r'$\cos\theta_2$',
    'phi_12': r'$\Delta\phi$',
    'phi_jl': r'$\phi_{JL}$',
    'luminosity_distance': '$d_L$',
    'dec': r'$\mathrm{DEC}$',
    'ra': r'$\mathrm{RA}$',
    'iota': r'$\iota$',
    'cos_iota': r'$\cos\iota$',
    'theta_jn': r'$\theta_{JN}$',
    'cos_theta_jn': r'$\cos\theta_{JN}$',
    'psi': r'$\psi$',
    'phase': r'$\phi$',
    'geocent_time': '$t_c$',
    'time_jitter': '$t_j$',
    'lambda_1': r'$\Lambda_1$',
    'lambda_2': r'$\Lambda_2$',
    'lambda_tilde': r'$\tilde{\Lambda}$',
    'delta_lambda_tilde': r'$\delta\tilde{\Lambda}$',
    'chi_1': r'$\chi_1$',
    'chi_2': r'$\chi_2$',
    'chi_1_in_plane': r'$\chi_{1, \perp}$',
    'chi_2_in_plane': r'$\chi_{2, \perp}$',
    'dchi_Minus2': r'$\delta\varphi_{-2}$',
    'dchi_0': r'$\delta\varphi_{0}$',
    'dchi_1': r'$\delta\varphi_{1}$',
    'dchi_2': r'$\delta\varphi_{2}$',
    'dchi_3': r'$\delta\varphi_{3}$',
    'dchi_3S': r'$\delta\varphi_{3S}$',
    'dchi_3NS': r'$\delta\varphi_{3NS}$',
    'dchi_4': r'$\delta\varphi_{4}$',
    'dchi_4S': r'$\delta\varphi_{4S}$',
    'dchi_4NS': r'$\delta\varphi_{4NS}$',
    'dchi_5l': r'$\delta\varphi_{5l}$',
    'dchi_5lS': r'$\delta\varphi_{5lS}$',
    'dchi_5lNS': r'$\delta\varphi_{5lNS}$',
    'dchi_6': r'$\delta\varphi_{6}$',
    'dchi_6S': r'$\delta\varphi_{6S}$',
    'dchi_6NS': r'$\delta\varphi_{6NS}$',
    'dchi_6l': r'$\delta\varphi_{6l}$',
    'dchi_7': r'$\delta\varphi_{7}$',
    'dchi_7S': r'$\delta\varphi_{7S}$',
    'dchi_7NS': r'$\delta\varphi_{7NS}$',
    'dchi_kappaS': r'$\delta\kappa_s$',
    'dchi_kappaA': r'$\delta\kappa_a$',
    'dchi_kappa1': r'$\delta\kappa_1$',
    'dchi_kappa2': r'$\delta\kappa_2$',
    'f_window_div_f_peak': r'$f^\mathrm{tape} / f_{22}^\mathrm{peak}$',
    'NumCycles': r'$\mathcal{N}_\mathrm{GW}$',
}



import unittest

import bilby_tgr
import numpy as np


class TestPriorInstantiationWithoutOptionalPriors(unittest.TestCase):
    def setUp(self):
        """ This is called before each test: use to reduce repeated code """
        self.frequency_array = np.linspace(20, 2048, 1000)

    def tearDown(self):
        """ This is called after each test: use to clean up"""
        del self.frequency_array

    def test_lal_binary_black_hole_TIGER(self):
        """ A simple test of the TIGER_binary_black_hole """

        # Call the function with a fixed set of parameters
        polarization_dict = bilby_tgr.tiger.source.lal_binary_black_hole_TIGER_PhenomX(
            frequency_array=self.frequency_array,
            mass_1=31,
            mass_2=30,
            luminosity_distance=1000,
            a_1=0.1,
            tilt_1=0.2,
            phi_12=0.2,
            a_2=0.2,
            tilt_2=0.2,
            phi_jl=0.2,
            theta_jn=1.2,
            phase=0.4,
            dchi_minus2=0.1,
            dchi_minus1=0.1,
            dchi_0=0.1,
            dchi_1=0.1,
            dchi_2=0.1,
            dchi_3=0.1,
            dchi_4=0.1,
            dchi_5l=0.1,
            dchi_6=0.1,
            dchi_6l=0.1,
            dchi_7=0.1,
            db_1=0.1,
            db_2=0.1,
            db_3=0.1,
            db_4=0.1,
            dc_1=0.1,
            dc_2=0.1,
            dc_4=0.1,
            dc_l=0.1,
            waveform_approximant='IMRPhenomXPHM')

        # Run tests to check the output is as expected: a dictionary of arrays
        self.assertTrue(isinstance(polarization_dict, dict))
        self.assertEqual(list(polarization_dict.keys()), ["plus", "cross"])
        self.assertTrue(isinstance(polarization_dict["plus"], np.ndarray))
        self.assertTrue(isinstance(polarization_dict["cross"], np.ndarray))

    def test_lal_binary_neutron_star_TIGER(self):
        """ A simple test of the TIGER_binary_black_hole """

        # Call the function with a fixed set of parameters
        polarization_dict = bilby_tgr.tiger.source.lal_binary_neutron_star_TIGER_PhenomX(
            frequency_array=self.frequency_array,
            mass_1=1.4,
            mass_2=1.4,
            luminosity_distance=100,
            a_1=0.1,
            tilt_1=0.2,
            phi_12=0.2,
            a_2=0.05,
            tilt_2=0.2,
            phi_jl=0.2,
            theta_jn=1.2,
            phase=0.4,
            lambda_1=500,
            lambda_2=500,
            dchi_minus2=0.1,
            dchi_minus1=0.1,
            dchi_0=0.1,
            dchi_1=0.1,
            dchi_2=0.1,
            dchi_3=0.1,
            dchi_4=0.1,
            dchi_5l=0.1,
            dchi_6=0.1,
            dchi_6l=0.1,
            dchi_7=0.1,
            waveform_approximant='IMRPhenomXP_NRTidalv2')

        # Run tests to check the output is as expected: a dictionary of arrays
        self.assertTrue(isinstance(polarization_dict, dict))
        self.assertEqual(list(polarization_dict.keys()), ["plus", "cross"])
        self.assertTrue(isinstance(polarization_dict["plus"], np.ndarray))
        self.assertTrue(isinstance(polarization_dict["cross"], np.ndarray))

    def test_lal_binary_black_hole_FTI(self):
        """ A simple test of the FTI binary_black_hole """

        # Call the function with a fixed set of parameters
        polarization_dict = bilby_tgr.fti.source.lal_binary_black_hole(
            frequency_array=self.frequency_array,
            mass_1=31,
            mass_2=30,
            luminosity_distance=1000,
            a_1=0.1,
            tilt_1=0.,
            phi_12=0.,
            a_2=0.2,
            tilt_2=0.,
            phi_jl=0.,
            theta_jn=1.2,
            phase=0.4,
            dchi_Minus2=0.1,
            dchi_0=0.1,
            dchi_1=0.1,
            dchi_2=0.1,
            dchi_3=0.1,
            dchi_3S=0.1,
            dchi_3NS=0.1,
            dchi_4=0.1,
            dchi_4S=0.1,
            dchi_4NS=0.1,
            dchi_5l=0.1,
            dchi_5lS=0.1,
            dchi_5lNS=0.1,
            dchi_6=0.1,
            dchi_6S=0.1,
            dchi_6NS=0.1,
            dchi_6l=0.1,
            dchi_7=0.1,
            dchi_7S=0.1,
            dchi_7NS=0.1,
            dchi_kappaS=0.1,
            dchi_kappaA=0.1,
            f_window_div_f_peak=1.0,
            NumCycles=1,
            waveform_approximant='IMRPhenomXAS')

        # Run tests to check the output is as expected: a dictionary of arrays
        self.assertTrue(isinstance(polarization_dict, dict))
        self.assertEqual(list(polarization_dict.keys()), ["plus", "cross"])
        self.assertTrue(isinstance(polarization_dict["plus"], np.ndarray))
        self.assertTrue(isinstance(polarization_dict["cross"], np.ndarray))

    def test_lal_binary_neutron_star_FTI(self):
        """ A simple test of the FTI binary_neutron_star """

        # Call the function with a fixed set of parameters
        polarization_dict = bilby_tgr.fti.source.lal_binary_neutron_star(
            frequency_array=self.frequency_array,
            mass_1=1.4,
            mass_2=1.4,
            luminosity_distance=100,
            a_1=0.1,
            tilt_1=0.,
            phi_12=0.,
            a_2=0.05,
            tilt_2=0.,
            phi_jl=0.,
            theta_jn=1.2,
            phase=0.4,
            lambda_1=500,
            lambda_2=500,
            dchi_Minus2=0.1,
            dchi_0=0.1,
            dchi_1=0.1,
            dchi_2=0.1,
            dchi_3=0.1,
            dchi_3S=0.1,
            dchi_3NS=0.1,
            dchi_4=0.1,
            dchi_4S=0.1,
            dchi_4NS=0.1,
            dchi_5l=0.1,
            dchi_5lS=0.1,
            dchi_5lNS=0.1,
            dchi_6=0.1,
            dchi_6S=0.1,
            dchi_6NS=0.1,
            dchi_6l=0.1,
            dchi_7=0.1,
            dchi_7S=0.1,
            dchi_7NS=0.1,
            dchi_kappaS=0.1,
            dchi_kappaA=0.1,
            f_window_div_f_peak=1.0,
            NumCycles=1,
            waveform_approximant='IMRPhenomXAS_NRTidalv2')

        # Run tests to check the output is as expected: a dictionary of arrays
        self.assertTrue(isinstance(polarization_dict, dict))
        self.assertEqual(list(polarization_dict.keys()), ["plus", "cross"])
        self.assertTrue(isinstance(polarization_dict["plus"], np.ndarray))
        self.assertTrue(isinstance(polarization_dict["cross"], np.ndarray))


if __name__ == "__main__":
    unittest.main()



from lal import CreateDict
import lalsimulation as ls
import lal
import numpy as np
from lal import MSUN_SI

# This gives polarizations from a given hlm mode by adding both the positive and negative m.
# see eqs. (5a) and (5b) of https://arxiv.org/pdf/2203.13937.pdf
# Directly coded from LALSimulation. see here
# https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimSphHarmMode.c#L175


def AddmodesFD(hlmf, L, M, iota, phiRef):
    hlmf = hlmf.data.data
    Y = lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phiRef, -2, L, M)
    Ymstar = np.conj(lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phiRef, -2, L, -M))
    factorp = 0.5 * (Y + (-1)**(L) * Ymstar)
    factorc = 1j * 0.5 * (Y - (-1)**(L) * Ymstar)
    hpf = factorp * hlmf
    hcf = factorc * hlmf
    return hpf, hcf


def create_dict_with_nongr_params(
        dchi_Minus2, dchi_0, dchi_1, dchi_2, dchi_3, dchi_3S, dchi_3NS, dchi_4, dchi_4S, dchi_4NS,
        dchi_5l, dchi_5lS, dchi_5lNS, dchi_6, dchi_6S, dchi_6NS, dchi_6l, dchi_7, dchi_7S, dchi_7NS,
        dchi_kappaS, dchi_kappaA):

    nongrParams = lal.CreateDict()

    if dchi_Minus2 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchiMinus2", dchi_Minus2)
    if dchi_0 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi0", dchi_0)
    if dchi_1 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi1", dchi_1)
    if dchi_2 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi2", dchi_2)
    if dchi_3 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi3", dchi_3)
    if dchi_3S != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi3S", dchi_3S)
    if dchi_3NS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi3NS", dchi_3NS)
    if dchi_4 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi4", dchi_4)
    if dchi_4S != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi4S", dchi_4S)
    if dchi_4NS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi4NS", dchi_4NS)
    if dchi_5l != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi5l", dchi_5l)
    if dchi_5lS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi5lS", dchi_5lS)
    if dchi_5lNS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi5lNS", dchi_5lNS)
    if dchi_6 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi6", dchi_6)
    if dchi_6S != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi6S", dchi_6S)
    if dchi_6NS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi6NS", dchi_6NS)
    if dchi_6l != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi6l", dchi_6l)
    if dchi_7 != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi7", dchi_7)
    if dchi_7S != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi7S", dchi_7S)
    if dchi_7NS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchi7NS", dchi_7NS)
    if dchi_kappaS != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchikappaS", dchi_kappaS)
    if dchi_kappaA != 0.:
        lal.DictInsertREAL8Value(nongrParams, "dchikappaA", dchi_kappaA)

    return nongrParams


def lal_binary_black_hole(
        frequency_array, mass_1, mass_2, luminosity_distance,
        a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl, theta_jn, phase,
        dchi_Minus2=0, dchi_0=0, dchi_1=0, dchi_2=0, dchi_3=0, dchi_3S=0, dchi_3NS=0,
        dchi_4=0, dchi_4S=0, dchi_4NS=0, dchi_5l=0, dchi_5lS=0, dchi_5lNS=0,
        dchi_6=0, dchi_6S=0, dchi_6NS=0, dchi_6l=0, dchi_7=0, dchi_7S=0, dchi_7NS=0,
        dchi_kappaS=0, dchi_kappaA=0, f_window_div_f_peak=1, NumCycles=1,
        **kwargs):

    """
    A binary black hole waveform using lalsimulation with
    parameterized corrections in the FTI framework.

    Parameters
    ----------
    frequency_array: array_like
        The frequencies at which we want to calculate the strain
    mass_1: float
        The mass of the heavier object in solar masses
    mass_2: float
        The mass of the lighter object in solar masses
    luminosity_distance: float
        The luminosity distance in megaparsec
    a_1: float
        Dimensionless primary spin magnitude
    tilt_1: float
        Primary tilt angle
    phi_12: float
        Azimuthal angle between the component spins
    a_2: float
        Dimensionless secondary spin magnitude
    tilt_2: float
        Secondary tilt angle
    phi_jl: float
        Azimuthal angle between the total and orbital angular momenta
    theta_jn: float
        Orbital inclination
    phase: float
        The phase at coalescence
    dchi_Minus2: float
        The non-GR inspiral parameter at -1PN
    dchi_0: float
        The non-GR inspiral parameter at 0PN
    dchi_1: float
        The non-GR inspiral parameter at 0.5PN
    dchi_2: float
        The non-GR inspiral parameter at 1PN
    dchi_3: float
        The non-GR inspiral parameter at 1.5PN
    dchi_3S: float
        The non-GR inspiral parameter at 1.5PN, spin part
    dchi_3NS: float
        The non-GR inspiral parameter at 1.5PN, non-spinning part
    dchi_4: float
        The non-GR inspiral parameter at 2PN
    dchi_4S: float
        The non-GR inspiral parameter at 2PN, spin part
    dchi_4NS: float
        The non-GR inspiral parameter at 2PN, non-spinning part
    dchi_5l: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction
    dchi_5lS: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction, spin part
    dchi_5lNS: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction, non-spinning part
    dchi_6: float
        The non-GR inspiral parameter at 3PN
    dchi_6S: float
        The non-GR inspiral parameter at 3PN, spin part
    dchi_6NS: float
        The non-GR inspiral parameter at 3PN, non-spinning part
    dchi_6l: float
        The non-GR inspiral parameter at 3PN, logarithmic correction
    dchi_7: float
        The non-GR inspiral parameter at 3.5PN
    dchi_7S: float
        The non-GR inspiral parameter at 3.5PN, spin part
    dchi_7NS: float
        The non-GR inspiral parameter at 3.5PN, non-spinning part
    dchi_kappaS: float
        The non-GR inspiral parameter for the symmetric
        combination of the spin-induced quadrupole moments
    dchi_kappaA: float
        The non-GR inspiral parameter for the anti-symmetric
        combination of the spin-induced quadrupole moments
    f_window_div_f_peak: float
        The tapering frequency as fraction of the peak frequency of the 22-mode
    NumCycles: float
        The number of gravitational wave cycles over which the tapering happens
    kwargs: dict
        Optional keyword arguments

    Returns
    -------
    dict: A dictionary with the plus and cross polarisation strain modes
    """

    waveform_kwargs = dict(
        waveform_approximant='IMRPhenomXAS', reference_frequency=50.0,
        minimum_frequency=20.0, maximum_frequency=frequency_array[-1],
        catch_waveform_errors=False, pn_spin_order=-1, pn_tidal_order=-1,
        pn_phase_order=-1, pn_amplitude_order=0, n_modes=4)
    waveform_kwargs.update(kwargs)

    # Create LALDict with FTI parameters
    nongrParams = create_dict_with_nongr_params(
        dchi_Minus2, dchi_0, dchi_1, dchi_2, dchi_3, dchi_3S, dchi_3NS, dchi_4, dchi_4S, dchi_4NS,
        dchi_5l, dchi_5lS, dchi_5lNS, dchi_6, dchi_6S, dchi_6NS, dchi_6l, dchi_7, dchi_7S, dchi_7NS,
        dchi_kappaS, dchi_kappaA)

    fRef = waveform_kwargs['reference_frequency']
    fMin = waveform_kwargs['minimum_frequency']
    fMax = frequency_array[-1]
    deltaF = frequency_array[1] - frequency_array[0]
    approximant = waveform_kwargs['waveform_approximant']
    nmodes = waveform_kwargs['n_modes']

    m1_SI = mass_1 * MSUN_SI
    m2_SI = mass_2 * MSUN_SI

    try:
        iota, S1x, S1y, S1z, S2x, S2y, S2z = ls.SimInspiralTransformPrecessingNewInitialConditions(
            theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2, m1_SI, m2_SI, fRef, phase)
    except Exception as e:
        if waveform_kwargs["catch_waveform_errors"]:
            return None
        else:
            raise Exception(e)

    dL = luminosity_distance * 1.0e6 * ls.lal.PC_SI

    # Treat waveform models without higher modes separately
    if approximant in ['SEOBNRv4_ROM', 'SEOBNRv5_ROM', 'IMRPhenomXAS']:
        approx = ls.GetApproximantFromString(approximant)

        try:
            hpf, hcf = ls.SimInspiralChooseFDWaveform(
                m1_SI, m2_SI, 0, 0, S1z, 0, 0, S2z, dL, iota, phase, 0, 0, 0, deltaF, fMin, fMax, fRef, None, approx)
        except Exception as e:
            if waveform_kwargs["catch_waveform_errors"]:
                return None
            else:
                raise Exception(e)

        ls.SimInspiralTestingGRCorrections(
            hpf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)
        ls.SimInspiralTestingGRCorrections(
            hcf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)

        h_plus = hpf.data.data
        h_cross = hcf.data.data

    elif approximant == 'SEOBNRv4HM_ROM':

        if nmodes == 1:
            approx = ls.GetApproximantFromString('SEOBNRv4_ROM')

            try:
                hpf, hcf = ls.SimInspiralChooseFDWaveform(
                    m1_SI, m2_SI, 0, 0, S1z, 0, 0, S2z, dL, iota, phase, 0, 0, 0,
                    deltaF, fMin, fMax, fRef, None, approx)
            except Exception as e:
                if waveform_kwargs["catch_waveform_errors"]:
                    return None
                else:
                    raise Exception(e)

            ls.SimInspiralTestingGRCorrections(
                hpf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)
            ls.SimInspiralTestingGRCorrections(
                hcf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)

            h_plus = hpf.data.data
            h_cross = hcf.data.data

        elif nmodes in [2, 3, 4, 5]:

            try:
                hlm_LAL = ls.SimIMRSEOBNRv4HMROM_Modes(
                    phase, deltaF, fMin, fMax, fRef, dL, m1_SI, m2_SI, S1z, S2z, -1, nmodes, use_hybridization=True)
            except Exception as e:
                if waveform_kwargs["catch_waveform_errors"]:
                    return None
                else:
                    raise Exception(e)

            h_plus = None
            h_cross = None

            modes = [[2, 2], [3, 3], [2, 1], [4, 4], [5, 5]][:nmodes]
            for L, M in modes:
                h = ls.SphHarmFrequencySeriesGetMode(hlm_LAL, L, -M)

                ls.SimInspiralTestingGRCorrections(
                    h, L, M, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)
                hp_corr_data, hc_corr_data = AddmodesFD(h, L, -M, iota, phase)

                if h_plus is None:
                    h_plus = hp_corr_data
                    h_cross = hc_corr_data
                else:
                    h_plus += hp_corr_data
                    h_cross += hc_corr_data

        else:
            raise Exception("The number of modes used with SEOBNRv4HM_ROM cannot be n_modes = {}.".format(nmodes))

    else:
        raise Exception("Waveform approximant {} is not supported. \
                        Use SEOBNRv4_ROM, SEOBNRv5_ROM, IMRPhenomXAS, \
                        or SEOBNRv4HM_ROM instead.".format(approximant))

    return dict(plus=h_plus, cross=h_cross)


def lal_binary_neutron_star(
        frequency_array, mass_1, mass_2, luminosity_distance,
        a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl, theta_jn, phase, lambda_1, lambda_2,
        dchi_Minus2=0, dchi_0=0, dchi_1=0, dchi_2=0, dchi_3=0, dchi_3S=0, dchi_3NS=0,
        dchi_4=0, dchi_4S=0, dchi_4NS=0, dchi_5l=0, dchi_5lS=0, dchi_5lNS=0,
        dchi_6=0, dchi_6S=0, dchi_6NS=0, dchi_6l=0, dchi_7=0, dchi_7S=0, dchi_7NS=0,
        dchi_kappaS=0, dchi_kappaA=0, f_window_div_f_peak=1, NumCycles=1,
        **kwargs):

    """
    A binary neutron star waveform using lalsimulation with
    parameterized corrections in the FTI framework.

    Parameters
    ----------
    frequency_array: array_like
        The frequencies at which we want to calculate the strain
    mass_1: float
        The mass of the heavier object in solar masses
    mass_2: float
        The mass of the lighter object in solar masses
    luminosity_distance: float
        The luminosity distance in megaparsec
    a_1: float
        Dimensionless primary spin magnitude
    tilt_1: float
        Primary tilt angle
    phi_12: float
        Azimuthal angle between the component spins
    a_2: float
        Dimensionless secondary spin magnitude
    tilt_2: float
        Secondary tilt angle
    phi_jl: float
        Azimuthal angle between the total and orbital angular momenta
    theta_jn: float
        Orbital inclination
    phase: float
        The phase at coalescence
    lambda_1: float
        The tides of the heavier object
    lambda_2: float
        The tides of the lighter object
    dchi_Minus2: float
        The non-GR inspiral parameter at -1PN
    dchi_0: float
        The non-GR inspiral parameter at 0PN
    dchi_1: float
        The non-GR inspiral parameter at 0.5PN
    dchi_2: float
        The non-GR inspiral parameter at 1PN
    dchi_3: float
        The non-GR inspiral parameter at 1.5PN
    dchi_3S: float
        The non-GR inspiral parameter at 1.5PN, spin part
    dchi_3NS: float
        The non-GR inspiral parameter at 1.5PN, non-spinning part
    dchi_4: float
        The non-GR inspiral parameter at 2PN
    dchi_4S: float
        The non-GR inspiral parameter at 2PN, spin part
    dchi_4NS: float
        The non-GR inspiral parameter at 2PN, non-spinning part
    dchi_5l: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction
    dchi_5lS: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction, spin part
    dchi_5lNS: float
        The non-GR inspiral parameter at 2.5PN, logarithmic correction, non-spinning part
    dchi_6: float
        The non-GR inspiral parameter at 3PN
    dchi_6S: float
        The non-GR inspiral parameter at 3PN, spin part
    dchi_6NS: float
        The non-GR inspiral parameter at 3PN, non-spinning part
    dchi_6l: float
        The non-GR inspiral parameter at 3PN, logarithmic correction
    dchi_7: float
        The non-GR inspiral parameter at 3.5PN
    dchi_7S: float
        The non-GR inspiral parameter at 3.5PN, spin part
    dchi_7NS: float
        The non-GR inspiral parameter at 3.5PN, non-spinning part
    dchi_kappaS: float
        The non-GR inspiral parameter for the symmetric
        combination of the spin-induced quadrupole moments
    dchi_kappaA: float
        The non-GR inspiral parameter for the anti-symmetric
        combination of the spin-induced quadrupole moments
    f_window_div_f_peak: float
        The tapering frequency as fraction of the peak frequency of the 22-mode
    NumCycles: float
        The number of gravitational wave cycles over which the tapering happens
    kwargs: dict
        Optional keyword arguments

    Returns
    -------
    dict: A dictionary with the plus and cross polarisation strain modes
    """

    waveform_kwargs = dict(
        waveform_approximant='IMRPhenomXAS_NRTidalv2', reference_frequency=50.0,
        minimum_frequency=20.0, maximum_frequency=frequency_array[-1],
        catch_waveform_errors=False, pn_spin_order=-1, pn_tidal_order=-1,
        pn_phase_order=-1, pn_amplitude_order=0)
    waveform_kwargs.update(kwargs)

    # If no waveform dictionary is passed, create one
    wf_dict = waveform_kwargs.get("lal_waveform_dictionary", CreateDict())
    # Update waveform dictionary with tidal parameters
    ls.SimInspiralWaveformParamsInsertTidalLambda1(wf_dict, float(lambda_1))
    ls.SimInspiralWaveformParamsInsertTidalLambda2(wf_dict, float(lambda_2))

    # Create LALDict with FTI parameters
    nongrParams = create_dict_with_nongr_params(
        dchi_Minus2, dchi_0, dchi_1, dchi_2, dchi_3, dchi_3S, dchi_3NS, dchi_4, dchi_4S, dchi_4NS,
        dchi_5l, dchi_5lS, dchi_5lNS, dchi_6, dchi_6S, dchi_6NS, dchi_6l, dchi_7, dchi_7S, dchi_7NS,
        dchi_kappaS, dchi_kappaA)

    fRef = waveform_kwargs['reference_frequency']
    fMin = waveform_kwargs['minimum_frequency']
    fMax = frequency_array[-1]
    deltaF = frequency_array[1] - frequency_array[0]
    approximant = waveform_kwargs['waveform_approximant']

    m1_SI = mass_1 * MSUN_SI
    m2_SI = mass_2 * MSUN_SI

    try:
        iota, S1x, S1y, S1z, S2x, S2y, S2z = ls.SimInspiralTransformPrecessingNewInitialConditions(
            theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2, m1_SI, m2_SI, fRef, phase)
    except Exception as e:
        if waveform_kwargs["catch_waveform_errors"]:
            return None
        else:
            raise Exception(e)

    dL = luminosity_distance * 1.0e6 * ls.lal.PC_SI

    if approximant in ['SEOBNRv4_ROM_NRTidalv2_NSBH', 'SEOBNRv4_ROM_NRTidalv2',
                       'SEOBNRv4T_surrogate', 'IMRPhenomXAS_NRTidalv2']:
        approx = ls.GetApproximantFromString(approximant)

        try:
            hpf, hcf = ls.SimInspiralChooseFDWaveform(
                m1_SI, m2_SI, 0, 0, S1z, 0, 0, S2z, dL, iota, phase, 0, 0, 0, deltaF, fMin, fMax, fRef, wf_dict, approx)
        except Exception as e:
            if waveform_kwargs["catch_waveform_errors"]:
                return None
            else:
                raise Exception(e)

        ls.SimInspiralTestingGRCorrections(
            hpf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)
        ls.SimInspiralTestingGRCorrections(
            hcf, 2, 2, m1_SI, m2_SI, S1z, S2z, fMin, fRef, f_window_div_f_peak, NumCycles, nongrParams)

        h_plus = hpf.data.data
        h_cross = hcf.data.data

    else:
        raise Exception("Waveform approximant {} is not supported. \
                        Use SEOBNRv4_ROM_NRTidalv2_NSBH, SEOBNRv4_ROM_NRTidalv2, 'IMRPhenomXAS_NRTidalv2', \
                        or SEOBNRv4T_surrogate instead.".format(approximant))

    return dict(plus=h_plus, cross=h_cross)



import lalsimulation
import numpy as np
from bilby.core import utils
from bilby.core.utils import logger
from bilby.gw.conversion import bilby_to_lalsimulation_spins
from bilby.gw.source import _base_lal_cbc_fd_waveform
from lal import CreateDict


def lal_binary_black_hole(
    frequency_array, mass_1, mass_2, luminosity_distance,
    a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl,
    theta_jn, phase,
    domega220, dtau220, domega330, dtau330, domega210, dtau210, domega440, dtau440, domega550, dtau550,
    **kwargs
):
    """
    A binary black hole waveform using lalsimulation with
    parameterized corrections in the TIGER framework.

    Parameters
    ----------
    frequency_array: array_like
        The frequencies at which we want to calculate the strain
    mass_1: float
        The mass of the heavier object in solar masses
    mass_2: float
        The mass of the lighter object in solar masses
    luminosity_distance: float
        The luminosity distance in megaparsec
    a_1: float
        Dimensionless primary spin magnitude
    tilt_1: float
        Primary tilt angle
    phi_12: float
        Azimuthal angle between the component spins
    a_2: float
        Dimensionless secondary spin magnitude
    tilt_2: float
        Secondary tilt angle
    phi_jl: float
        Azimuthal angle between the total and orbital angular momenta
    theta_jn: float
        Orbital inclination
    phase: float
        The phase at coalescence
    domega220: float
        The non-GR fractional deviation to the 220 frequency
    dtau220: float
        The non-GR fractional deviation to the 220 damping time
    domega330: float
        The non-GR fractional deviation to the 330 frequency
    dtau330: float
        The non-GR fractional deviation to the 330 damping time
    domega210: float
        The non-GR fractional deviation to the 210 frequency
    dtau210: float
        The non-GR fractional deviation to the 210 damping time
    domega440: float
        The non-GR fractional deviation to the 440 frequency
    dtau440: float
        The non-GR fractional deviation to the 440 damping time
    domega550: float
        The non-GR fractional deviation to the 550 frequency
    dtau550: float
        The non-GR fractional deviation to the 550 damping time
    kwargs: dict
        Optional keyword arguments

    Returns
    -------
    dict: A dictionary with the plus and cross polarisation strain modes
    """
    waveform_kwargs = dict(
        waveform_approximant='pSEOBNRv4HM_PA', reference_frequency=50.0,
        minimum_frequency=20.0, maximum_frequency=frequency_array[-1],
        catch_waveform_errors=False, pn_spin_order=-1, pn_tidal_order=-1,
        pn_phase_order=-1, pn_amplitude_order=0)
    waveform_kwargs.update(kwargs)

    # If no waveform dictionary is passed, create one
    wf_dict = waveform_kwargs.get("lal_waveform_dictionary", CreateDict())

    # Fractional deviation in the 220 frequency and damping time
    if domega220 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDOmega220(wf_dict, float(domega220))
    if dtau220 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDTau220(wf_dict, float(dtau220))

    # Fractional deviation in the 210 frequency and damping time
    if domega210 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDOmega210(wf_dict, float(domega210))
    if dtau210 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDTau210(wf_dict, float(dtau210))

    # Fractional deviation in the 330 frequency and damping time
    if domega330 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDOmega330(wf_dict, float(domega330))
    if dtau330 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDTau330(wf_dict, float(dtau330))

    # Fractional deviation in the 440 frequency and damping time
    if domega440 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDOmega440(wf_dict, float(domega440))
    if dtau440 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDTau440(wf_dict, float(dtau440))

    # Fractional deviation in the 550 frequency and damping time
    if domega550 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDOmega550(wf_dict, float(domega550))
    if dtau550 != 0.:
        lalsimulation.SimInspiralWaveformParamsInsertDTau550(wf_dict, float(dtau550))

    # Update waveform dictionary with TIGER coefficients
    waveform_kwargs["lal_waveform_dictionary"] = wf_dict

    return _base_lal_cbc_fd_waveform(
        frequency_array=frequency_array, mass_1=mass_1, mass_2=mass_2,
        luminosity_distance=luminosity_distance, theta_jn=theta_jn, phase=phase,
        a_1=a_1, a_2=a_2, tilt_1=tilt_1, tilt_2=tilt_2, phi_12=phi_12,
        phi_jl=phi_jl, **waveform_kwargs)


def gwsignal_binary_black_hole(frequency_array, mass_1, mass_2, luminosity_distance, a_1, tilt_1,
                               phi_12, a_2, tilt_2, phi_jl, theta_jn, phase,
                               domega220, dtau220, domega330, dtau330, domega210, dtau210, domega440,
                               dtau440, domega550, dtau550, domega320, dtau320, domega430, dtau430,
                               dA22, dw22, dA33, dw33, dA21, dw21, dA44, dw44, dA55, dw55, dA32, dw32, dA43, dw43,
                               dTpeak, da6, ddSO, **kwargs):
    """
    A parametrized SEOB binary black hole waveform model using GWsignal

    Parameters
    ==========
    frequency_array: array_like
        The frequencies at which we want to calculate the strain
    mass_1: float
        The mass of the heavier object in solar masses
    mass_2: float
        The mass of the lighter object in solar masses
    luminosity_distance: float
        The luminosity distance in megaparsec
    a_1: float
        Dimensionless primary spin magnitude
    tilt_1: float
        Primary tilt angle
    phi_12: float
        Azimuthal angle between the two component spins
    a_2: float
        Dimensionless secondary spin magnitude
    tilt_2: float
        Secondary tilt angle
    phi_jl: float
        Azimuthal angle between the total binary angular momentum and the
        orbital angular momentum
    theta_jn: float
        Angle between the total binary angular momentum and the line of sight
    phase: float
        The phase at coalescence
    domega220: float
        The non-GR fractional deviation to the 220 frequency
    dtau220: float
        The non-GR fractional deviation to the 220 damping time
    domega330: float
        The non-GR fractional deviation to the 330 frequency
    dtau330: float
        The non-GR fractional deviation to the 330 damping time
    domega210: float
        The non-GR fractional deviation to the 210 frequency
    dtau210: float
        The non-GR fractional deviation to the 210 damping time
    domega440: float
        The non-GR fractional deviation to the 440 frequency
    dtau440: float
        The non-GR fractional deviation to the 440 damping time
    domega550: float
        The non-GR fractional deviation to the 550 frequency
    dtau550: float
        The non-GR fractional deviation to the 550 damping time
    domega320: float
        The non-GR fractional deviation to the 320 frequency
    dtau320: float
        The non-GR fractional deviation to the 320 damping time
    domega430: float
        The non-GR fractional deviation to the 430 frequency
    dtau430: float
        The non-GR fractional deviation to the 430 damping time
    dA22: float
        The non-GR fractional deviation to the 22 merger amplitude
    dw22: float
        The non-GR fractional deviation to the 22 merger frequency
    dA33: float
        The non-GR fractional deviation to the 33 merger amplitude
    dw33: float
        The non-GR fractional deviation to the 33 merger frequency
    dA21: float
        The non-GR fractional deviation to the 21 merger amplitude
    dw21: float
        The non-GR fractional deviation to the 21 merger frequency
    dA44: float
        The non-GR fractional deviation to the 44 merger amplitude
    dw44: float
        The non-GR fractional deviation to the 44 merger frequency
    dA55: float
        The non-GR fractional deviation to the 55 merger amplitude
    dw55: float
        The non-GR fractional deviation to the 55 merger frequency
    dA32: float
        The non-GR fractional deviation to the 32 merger amplitude
    dw32: float
        The non-GR fractional deviation to the 32 merger frequency
    dA43: float
        The non-GR fractional deviation to the 43 merger amplitude
    dw43: float
        The non-GR fractional deviation to the 43 merger frequency
    dTpeak: float
        The non-GR additive deviation to the amplitude's peak time
    da6: float
        The non-GR additive deviation to the a6 calibration parameter
    ddSO: float
        The non-GR additive deviation to the dSO calibration parameter
    kwargs: dict
        Optional keyword arguments
        Supported arguments:

        - waveform_approximant
        - reference_frequency
        - minimum_frequency
        - maximum_frequency
        - catch_waveform_errors
        - pn_amplitude_order
        - mode_array:
          Activate a specific mode array and evaluate the model using those
          modes only.  e.g. waveform_arguments =
          dict(waveform_approximant='IMRPhenomHM', mode_array=[[2,2],[2,-2]])
          returns the 22 and 2-2 modes only of IMRPhenomHM.  You can only
          specify modes that are included in that particular model.  e.g.
          waveform_arguments = dict(waveform_approximant='IMRPhenomHM',
          mode_array=[[2,2],[2,-2],[5,5],[5,-5]]) is not allowed because the
          55 modes are not included in this model.  Be aware that some models
          only take positive modes and return the positive and the negative
          mode together, while others need to call both.  e.g.
          waveform_arguments = dict(waveform_approximant='IMRPhenomHM',
          mode_array=[[2,2],[4,-4]]) returns the 22 and 2-2 of IMRPhenomHM.
          However, waveform_arguments =
          dict(waveform_approximant='IMRPhenomXHM', mode_array=[[2,2],[4,-4]])
          returns the 22 and 4-4 of IMRPhenomXHM.

    Returns
    =======
    dict: A dictionary with the plus and cross polarisation strain modes

    Notes
    =====
    This function is a temporary wrapper to the interface that will
    likely be significantly changed or removed in a future release.
    This version is only intended to be used with `SEOBNRv5HM` and `SEOBNRv5PHM` and
    does not have full functionality for other waveform models.
    """

    from lalsimulation.gwsignal import GenerateFDWaveform
    from lalsimulation.gwsignal.models import gwsignal_get_waveform_generator
    import astropy.units as u

    waveform_kwargs = dict(
        waveform_approximant="SEOBNRv5PHM",
        reference_frequency=50.0,
        minimum_frequency=20.0,
        maximum_frequency=frequency_array[-1],
        catch_waveform_errors=False,
        mode_array=None,
        pn_amplitude_order=0,
    )
    waveform_kwargs.update(kwargs)

    waveform_approximant = waveform_kwargs['waveform_approximant']
    if waveform_approximant not in ["SEOBNRv5HM", "SEOBNRv5PHM"]:
        raise ValueError("The only supported approximants are SEOBNRv5PHM and SEOBNRv5HM.")

    reference_frequency = waveform_kwargs['reference_frequency']
    minimum_frequency = waveform_kwargs['minimum_frequency']
    maximum_frequency = waveform_kwargs['maximum_frequency']
    catch_waveform_errors = waveform_kwargs['catch_waveform_errors']
    mode_array = waveform_kwargs['mode_array']
    pn_amplitude_order = waveform_kwargs['pn_amplitude_order']

    if pn_amplitude_order != 0:
        # This is to mimic the behaviour in
        # https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L5542
        if pn_amplitude_order == -1:
            if waveform_approximant in ["SpinTaylorT4", "SpinTaylorT5"]:
                pn_amplitude_order = 3  # Equivalent to MAX_PRECESSING_AMP_PN_ORDER in LALSimulation
            else:
                pn_amplitude_order = 6  # Equivalent to MAX_NONPRECESSING_AMP_PN_ORDER in LALSimulation
        start_frequency = minimum_frequency * 2. / (pn_amplitude_order + 2)
    else:
        start_frequency = minimum_frequency

    # Call GWsignal generator
    wf_gen = gwsignal_get_waveform_generator(waveform_approximant)

    delta_frequency = frequency_array[1] - frequency_array[0]

    frequency_bounds = ((frequency_array >= minimum_frequency) *
                        (frequency_array <= maximum_frequency))

    iota, spin_1x, spin_1y, spin_1z, spin_2x, spin_2y, spin_2z = bilby_to_lalsimulation_spins(
        theta_jn=theta_jn, phi_jl=phi_jl, tilt_1=tilt_1, tilt_2=tilt_2,
        phi_12=phi_12, a_1=a_1, a_2=a_2, mass_1=mass_1 * utils.solar_mass, mass_2=mass_2 * utils.solar_mass,
        reference_frequency=reference_frequency, phase=phase)

    eccentricity = 0.0
    longitude_ascending_nodes = 0.0
    mean_per_ano = 0.0

    # Check if conditioning is needed
    condition = 0
    if wf_gen.metadata["implemented_domain"] == 'time':
        condition = 1

    # Create dict for gwsignal generator
    gwsignal_dict = {'mass1' : mass_1 * u.solMass,
                     'mass2' : mass_2 * u.solMass,
                     'spin1x' : spin_1x * u.dimensionless_unscaled,
                     'spin1y' : spin_1y * u.dimensionless_unscaled,
                     'spin1z' : spin_1z * u.dimensionless_unscaled,
                     'spin2x' : spin_2x * u.dimensionless_unscaled,
                     'spin2y' : spin_2y * u.dimensionless_unscaled,
                     'spin2z' : spin_2z * u.dimensionless_unscaled,
                     'deltaF' : delta_frequency * u.Hz,
                     'f22_start' : start_frequency * u.Hz,
                     'f_max': maximum_frequency * u.Hz,
                     'f22_ref': reference_frequency * u.Hz,
                     'phi_ref' : phase * u.rad,
                     'distance' : luminosity_distance * u.Mpc,
                     'inclination' : iota * u.rad,
                     'eccentricity' : eccentricity * u.dimensionless_unscaled,
                     'longAscNodes' : longitude_ascending_nodes * u.rad,
                     'meanPerAno' : mean_per_ano * u.rad,
                     # 'ModeArray': mode_array,
                     'condition': condition
                     }

    if mode_array is not None:
        gwsignal_dict.update(ModeArray=mode_array)

    if (
        domega220 != 0.0
        or domega330 != 0.0
        or domega210 != 0.0
        or domega440 != 0.0
        or domega550 != 0.0
        or domega320 != 0.0
        or domega430 != 0.0
    ):
        domega_dict = {'2,2': domega220,
                       '2,1': domega210,
                       '3,3': domega330,
                       '3,2': domega320,
                       '4,4': domega440,
                       '4,3': domega430,
                       '5,5': domega550,
                       }
        gwsignal_dict.update(domega_dict=domega_dict)

    if (
        dtau220 != 0.0
        or dtau330 != 0.0
        or dtau210 != 0.0
        or dtau440 != 0.0
        or dtau550 != 0.0
        or dtau320 != 0.0
        or dtau430 != 0.0
    ):
        dtau_dict = {'2,2': dtau220,
                     '2,1': dtau210,
                     '3,3': dtau330,
                     '3,2': dtau320,
                     '4,4': dtau440,
                     '4,3': dtau430,
                     '5,5': dtau550,
                     }
        gwsignal_dict.update(dtau_dict=dtau_dict)
    if (
        dA22 != 0.0
        or dA33 != 0.0
        or dA21 != 0.0
        or dA44 != 0.0
        or dA55 != 0.0
        or dA32 != 0.0
        or dA43 != 0.0
    ):
        dA_dict = {'2,2': dA22,
                   '2,1': dA21,
                   '3,3': dA33,
                   '3,2': dA32,
                   '4,4': dA44,
                   '4,3': dA43,
                   '5,5': dA55,
                   }
        gwsignal_dict.update(dA_dict=dA_dict)

    if (
        dw22 != 0.0
        or dw33 != 0.0
        or dw21 != 0.0
        or dw44 != 0.0
        or dw55 != 0.0
        or dw32 != 0.0
        or dw43 != 0.0
    ):
        dw_dict = {'2,2': dw22,
                   '2,1': dw21,
                   '3,3': dw33,
                   '3,2': dw32,
                   '4,4': dw44,
                   '4,3': dw43,
                   '5,5': dw55,
                   }
        gwsignal_dict.update(dw_dict=dw_dict)

    if dTpeak != 0.0:
        gwsignal_dict.update(dTpeak=dTpeak)

    if da6 != 0.0:
        gwsignal_dict.update(da6=da6)

    if ddSO != 0.0:
        gwsignal_dict.update(ddSO=ddSO)

    # Pass extra waveform arguments to gwsignal
    extra_args = waveform_kwargs.copy()

    for key in [
            "waveform_approximant",
            "reference_frequency",
            "minimum_frequency",
            "maximum_frequency",
            "catch_waveform_errors",
            "mode_array",
            "pn_spin_order",
            "pn_amplitude_order",
            "pn_tidal_order",
            "pn_phase_order",
            "numerical_relativity_file",
    ]:
        if key in extra_args.keys():
            del extra_args[key]

    gwsignal_dict.update(extra_args)

    try:
        hpc = GenerateFDWaveform(gwsignal_dict, wf_gen)
    except Exception as e:
        if not catch_waveform_errors:
            raise
        else:
            EDOM = (
                "Internal function call failed: Input domain error" in e.args[0]
            ) or "Input domain error" in e.args[
                0
            ]
            if EDOM:
                failed_parameters = dict(mass_1=mass_1, mass_2=mass_2,
                                         spin_1=(spin_1x, spin_1y, spin_1z),
                                         spin_2=(spin_2x, spin_2y, spin_2z),
                                         luminosity_distance=luminosity_distance,
                                         iota=iota, phase=phase,
                                         eccentricity=eccentricity,
                                         start_frequency=minimum_frequency)
                logger.warning("Evaluating the waveform failed with error: {}\n".format(e) +
                               "The parameters were {}\n".format(failed_parameters) +
                               "Likelihood will be set to -inf.")
                return None
            else:
                raise

    hplus = hpc.hp
    hcross = hpc.hc

    h_plus = np.zeros_like(frequency_array, dtype=complex)
    h_cross = np.zeros_like(frequency_array, dtype=complex)

    if len(hplus) > len(frequency_array):
        logger.debug("GWsignal waveform longer than bilby's `frequency_array`" +
                     "({} vs {}), ".format(len(hplus), len(frequency_array)) +
                     "probably because padded with zeros up to the next power of two length." +
                     " Truncating GWsignal array.")
        h_plus = hplus[:len(h_plus)]
        h_cross = hcross[:len(h_cross)]
    else:
        h_plus[:len(hplus)] = hplus
        h_cross[:len(hcross)] = hcross

    h_plus *= frequency_bounds
    h_cross *= frequency_bounds

    if condition:
        dt = 1 / hplus.df.value + hplus.epoch.value
        time_shift = np.exp(-1j * 2 * np.pi * dt * frequency_array[frequency_bounds])
        h_plus[frequency_bounds] *= time_shift
        h_cross[frequency_bounds] *= time_shift

    return dict(plus=h_plus, cross=h_cross)



import bilby.gw.prior as prior
import os
from bilby.core.utils import logger

DEFAULT_PRIOR_DIR = os.path.join(os.path.dirname(__file__), 'prior_files')


class BBHPriorDict(prior.BBHPriorDict):
    def __init__(self, dictionary=None, filename=None, aligned_spin=False,
                 conversion_function=None):
        """ Initialises a Prior set for Binary Black holes with SIQM parameter.
        Copied from bilby.

        Parameters
        ==========
        dictionary: dict, optional
            See superclass
        filename: str, optional
            See superclass
        conversion_function: func
            Function to convert between sampled parameters and constraints.
            By default this generates many additional parameters, see
            BBHPriorDict.default_conversion_function
        """
        if dictionary is None and filename is None:
            if aligned_spin:
                fname = 'aligned_spins_bbh.prior'
                logger.info('Using aligned spin prior')
            else:
                fname = 'precessing_spins_bbh.prior'
            filename = os.path.join(DEFAULT_PRIOR_DIR, fname)
            logger.info('No prior given, using default BBH priors in {}.'.format(filename))
        elif filename is not None:
            if not os.path.isfile(filename):
                filename = os.path.join(DEFAULT_PRIOR_DIR, filename)
        super(prior.BBHPriorDict, self).__init__(dictionary=dictionary, filename=filename,
                                                 conversion_function=conversion_function)



import unittest
from copy import copy
from unittest.mock import patch

import astropy.units as u
import bilby
import lal
import numpy as np
from gwpy.frequencyseries import FrequencySeries
from lalsimulation.gwsignal.models.pyseobnr_model import SEOBNRv5PHM as wf_gen

import bilby_tgr


class TestpSEOBNR(unittest.TestCase):
    def setUp(self):

        self.waveform_kwargs = {
            "waveform_approximant": "SEOBNRv5PHM",
            "mass_1": 60.0,
            "mass_2": 15.0,
            "luminosity_distance": 1429.88270378,
            "a_1": 0.6,
            "a_2": 0.4,
            "tilt_1": 1.4,
            "tilt_2": 0.1,
            "phi_12": 0.76,
            "phi_jl": 1.31,
            "theta_jn": 0.45541468,
            "phase": 0.4675536,
            "minimum_frequency": 15.0,
            "reference_frequency": 20.0,
            "catch_waveform_errors": True,
        }

        self.pseob_params = {
            "domega220",
            "dtau220",
            "domega330",
            "dtau330",
            "domega210",
            "dtau210",
            "domega440",
            "dtau440",
            "domega550",
            "dtau550",
            "domega320",
            "dtau320",
            "domega430",
            "dtau430",
            "dA22",
            "dw22",
            "dA33",
            "dw33",
            "dA21",
            "dw21",
            "dA44",
            "dw44",
            "dA55",
            "dw55",
            "dA32",
            "dw32",
            "dA43",
            "dw43",
            "dTpeak",
            "da6",
            "ddSO",
        }

        self.frequency_array = bilby.core.utils.create_frequency_series(2048, 4)
        self.bad_parameters = copy(self.waveform_kwargs)
        self.bad_parameters["mass_1"] = 300000.0  # this generates an input domain error

    def test_pseob_works_runs_valid_parameters(self):
        self.waveform_kwargs.update({_: 0.0 for _ in self.pseob_params})
        # note: currently if the pSEOB specific variables are all 0, they are removed
        # from the parameters passed to gwsignal.
        # This makes it possible to run those tests without the proper changes in gwsignal
        # and pyseobnr. This is not a desired property and one would need to activate
        # the end-to-end pyseobnr tests, simply by making it so that one of the values
        # is non-zero. This would be possible once the other MR are merged:
        # * https://git.ligo.org/lscsoft/lalsuite/-/merge_requests/2319
        # * https://git.ligo.org/waveforms/software/pyseobnr/-/merge_requests/57

        # example for activating the end-to-end tests from bilby_tgr:
        # self.waveform_kwargs["dTpeak"] = 0.1
        self.assertIsInstance(
            bilby_tgr.pseob.source.gwsignal_binary_black_hole(
                self.frequency_array, **self.waveform_kwargs
            ),
            dict,
        )

    def test_waveform_error_catching(self):
        erroneous_parameters = copy(self.bad_parameters)
        assert erroneous_parameters["catch_waveform_errors"]
        erroneous_parameters.update({_: 0.0 for _ in self.pseob_params})
        self.assertIsNone(
            bilby_tgr.pseob.source.gwsignal_binary_black_hole(
                self.frequency_array, **erroneous_parameters
            )
        )

    def test_waveform_error_raising(self):
        raise_error_parameters = copy(self.waveform_kwargs)
        raise_error_parameters.update(self.bad_parameters)
        raise_error_parameters.update({_: 0.0 for _ in self.pseob_params})
        raise_error_parameters["catch_waveform_errors"] = False
        with self.assertRaises(Exception):
            bilby_tgr.pseob.source.gwsignal_binary_black_hole(
                self.frequency_array, **raise_error_parameters
            )

    def test_report_errors_that_are_not_input_domain_errors(self):
        raise_error_parameters = copy(self.waveform_kwargs)
        raise_error_parameters["mass_1"] = -10
        raise_error_parameters.update({_: 0.0 for _ in self.pseob_params})
        # this should be reported even if we are catching the errors
        raise_error_parameters["catch_waveform_errors"] = True
        with self.assertRaises(Exception):
            bilby_tgr.pseob.source.gwsignal_binary_black_hole(
                self.frequency_array, **raise_error_parameters
            )

    @staticmethod
    def fake_fd_signal(hp, hc, deltaT=1 / 2048):
        # generates fake fd signal
        epoch = lal.LIGOTimeGPS(-300)

        hp_lal = lal.CreateREAL8TimeSeries(
            "hplus", epoch, 0, deltaT, lal.DimensionlessUnit, len(hp)
        )
        hc_lal = lal.CreateREAL8TimeSeries(
            "hcross",
            epoch,
            0,
            1 / 2048,
            lal.DimensionlessUnit,
            len(hp),
        )

        hp_lal.data.data = hp
        hc_lal.data.data = hc

        chirplen = hp_lal.data.length
        chirplen_exp = np.frexp(chirplen)
        chirplen = int(np.ldexp(1, chirplen_exp[1]))
        deltaF = 1.0 / (chirplen * deltaT)

        lal.ResizeREAL8TimeSeries(hp_lal, hp_lal.data.length - chirplen, chirplen)
        lal.ResizeREAL8TimeSeries(hc_lal, hc_lal.data.length - chirplen, chirplen)

        # FFT - Using LAL routines
        hptilde = lal.CreateCOMPLEX16FrequencySeries(
            "FD H_PLUS",
            hp_lal.epoch,
            0.0,
            deltaF,
            lal.DimensionlessUnit,
            int(chirplen / 2.0 + 1),
        )
        hctilde = lal.CreateCOMPLEX16FrequencySeries(
            "FD H_CROSS",
            hc_lal.epoch,
            0.0,
            deltaF,
            lal.DimensionlessUnit,
            int(chirplen / 2.0 + 1),
        )

        plan = lal.CreateForwardREAL8FFTPlan(chirplen, 0)
        lal.REAL8TimeFreqFFT(hctilde, hc_lal, plan)
        lal.REAL8TimeFreqFFT(hptilde, hp_lal, plan)

        frequencies = hptilde.deltaF * np.arange(hptilde.data.length)
        epoch = hptilde.epoch.gpsSeconds + hptilde.epoch.gpsNanoSeconds / 1e9
        return (
            FrequencySeries(
                hptilde.data.data,
                frequencies=frequencies,
                epoch=epoch,
                name="hp",
                unit=u.Unit("strain") * u.s,
            ),
            FrequencySeries(
                hctilde.data.data,
                frequencies=frequencies,
                epoch=epoch,
                name="hc",
                unit=u.Unit("strain") * u.s,
            ),
        )

    def test_pseob_specific_parameters(self):
        # here we test the behaviour of the function "gwsignal_binary_black_hole"
        # until the execution of the "generate_fd_waveform" in gwsignal. In particular
        # the actual generate_fd_waveform is not called and only the parameters passed to
        # the function are checked.
        # The test does not require gwsignal to support pSEOB parameters.

        hp = np.zeros(1000)
        hc = np.zeros(1000)

        return_val_fd_hptilde, return_val_fd_hctilde = self.fake_fd_signal(
            hp, hc, 1 / 2048
        )

        with patch.object(wf_gen, "generate_fd_waveform", autospec=True) as wgen_gen_fd:

            for current_param in self.pseob_params:

                wgen_gen_fd.return_value = return_val_fd_hptilde, return_val_fd_hctilde

                wgen_gen_fd.reset_mock()
                parameters = copy(self.waveform_kwargs)
                parameters.update({_: 0.0 for _ in self.pseob_params})
                parameters.update({current_param: 0.1})

                bilby_tgr.pseob.source.gwsignal_binary_black_hole(
                    self.frequency_array, **parameters
                )

                # checks generated_fd_waveform is called
                wgen_gen_fd.assert_called_once()

                # checks parameters are passed: this is done inside the
                self.assertIsInstance(wgen_gen_fd.call_args_list[0].args[0], wf_gen)

                # check if the currently modified parameter is dA22, dtau32 etc
                for parameter_type in "dA", "domega", "dw", "dtau":
                    if current_param.startswith(parameter_type):
                        param_to_check = f"{parameter_type}_dict"

                        # the trailing part dtau330 -> 330 -> 3,3
                        element_to_check = ",".join(
                            current_param[len(parameter_type) : len(parameter_type) + 2]
                        )

                        self.assertIn(
                            param_to_check, wgen_gen_fd.call_args_list[0].kwargs
                        )
                        self.assertIn(
                            element_to_check,
                            wgen_gen_fd.call_args_list[0].kwargs[param_to_check],
                        )
                        self.assertEqual(
                            wgen_gen_fd.call_args_list[0].kwargs[param_to_check][
                                element_to_check
                            ],
                            0.1,
                        )
                        break

                else:
                    if current_param in ("dTpeak", "da6", "ddSO"):
                        self.assertIn(
                            current_param, wgen_gen_fd.call_args_list[0].kwargs
                        )
                        self.assertEqual(
                            wgen_gen_fd.call_args_list[0].kwargs[current_param], 0.1
                        )
                    else:
                        self.fail(f"Parameter {current_param} not found!")



import bilby.gw.conversion as conversion
from bilby.core.utils import logger


def dkappa1_and_dkappa2_to_dkappaS_and_dkappaA(dchi_kappa1, dchi_kappa2):
    """
    Convert the spin-induced quadrupole moment of the individual objects to
    their symmetric and anti-symmetric combinations.

    Parameters
    ----------
    dchi_kappa1: float
        Spin-induced quadrupole moment of the heavier object
    dchi_kappa2: float
        Spin-induced quadrupole moment of the lighter object

    Return
    ------
    dchi_kappaS: float
        symmetric combination of individual spin-induced quadrupole moments
    dchi_kappaA: float
        anti-symmetric combination of individual spin-induced quadrupole moments
    """

    dchi_kappaS = (dchi_kappa1 + dchi_kappa2) / 2.
    dchi_kappaA = (dchi_kappa1 - dchi_kappa2) / 2.
    return dchi_kappaS, dchi_kappaA


def dkappaS_and_dkappaA_to_dkappa1_and_dkappa2(dchi_kappaS, dchi_kappaA):
    """
    Convert the symmetric and anti-symmetric combinations of spin-induced quadrupole moments to
    the spin-induced quadrupole moments of the individual objects.

    Parameters
    ----------
    dchi_kappaS: float
        symmetric combination of individual spin-induced quadrupole moments
    dchi_kappaA: float
        anti-symmetric combination of individual spin-induced quadrupole moments

    Return
    ------
    dchi_kappa1: float
        Spin-induced quadrupole moment of the heavier object
    dchi_kappa2: float
        Spin-induced quadrupole moment of the lighter object

    """

    dchi_kappa1 = dchi_kappaS + dchi_kappaA
    dchi_kappa2 = dchi_kappaS - dchi_kappaA
    return dchi_kappa1, dchi_kappa2


def generate_fti_parameters(sample, added_keys=[]):
    """
    Generate all SIQM parameters for FTI SIQM test:
        dchi_kappaS, dchi_kappaA, dchi_kappa1, dchi_kappa2,

    Parameters
    ==========
    sample: dict, pandas.DataFrame

    Returns
    =======
    output_sample: dict, pandas.DataFrame
        Updated sample
    """

    output_sample = sample.copy()

    if 'f_window_div_f_peak' not in output_sample.keys():
        output_sample['f_window_div_f_peak'] = 1.0
        added_keys = added_keys + ['f_window_div_f_peak']

    if 'NumCycles' not in output_sample.keys():
        output_sample['NumCycles'] = 1
        added_keys = added_keys + ['NumCycles']

    if not any([key in output_sample for key in
                ['dchi_kappaS', 'dchi_kappaA', 'dchi_kappa1', 'dchi_kappa2']]):
        return output_sample, added_keys

    if (('dchi_kappaS' in output_sample.keys() or 'dchi_kappaA' in output_sample.keys()) and
            ('dchi_kappa1' in output_sample.keys() or 'dchi_kappa2' in output_sample.keys())):
        logger.disabled = False
        logger.warning("Incompatible dkappas provided. This may lead to unexpected behaviour. \
                        Provide (dchi_kappaS and dchi_kappaA) xor (dchi_kappa1 and dchi_kappa2) instead.")

    if 'dchi_kappaS' not in output_sample.keys() and 'dchi_kappaA' not in output_sample.keys():
        if 'dchi_kappa1' not in output_sample.keys():
            output_sample['dchi_kappa1'] = 0
            added_keys = added_keys + ['dchi_kappa1']
        if 'dchi_kappa2' not in output_sample.keys():
            output_sample['dchi_kappa2'] = 0
            added_keys = added_keys + ['dchi_kappa2']
        output_sample['dchi_kappaS'], output_sample['dchi_kappaA'] =\
            dkappa1_and_dkappa2_to_dkappaS_and_dkappaA(
                output_sample['dchi_kappa1'], output_sample['dchi_kappa2'])
        added_keys = added_keys + ['dchi_kappaS', 'dchi_kappaA']

    if 'dchi_kappa1' not in output_sample.keys() and 'dchi_kappa2' not in output_sample.keys():
        if 'dchi_kappaS' not in output_sample.keys():
            output_sample['dchi_kappaS'] = 0
            added_keys = added_keys + ['dchi_kappaS']
        if 'dchi_kappaA' not in output_sample.keys():
            output_sample['dchi_kappaA'] = 0
            added_keys = added_keys + ['dchi_kappaA']
        output_sample['dchi_kappa1'], output_sample['dchi_kappa2'] =\
            dkappaS_and_dkappaA_to_dkappa1_and_dkappa2(
                output_sample['dchi_kappaS'], output_sample['dchi_kappaA'])
        added_keys = added_keys + ['dchi_kappa1', 'dchi_kappa2']

    return output_sample, added_keys


def convert_to_lal_binary_black_hole_parameters(parameters):
    """
    (Adapted from bilby)
    Convert parameters we have into parameters we need.

    This is defined by the parameters of bilby_tgr.fti.source.lal_binary_black_hole()

    Mass: mass_1, mass_2
    Spin: a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl
    Extrinsic: luminosity_distance, theta_jn, phase, ra, dec, geocent_time, psi
    FTI: dchi_Minus2, dchi_0, dchi_1, dchi_2, dchi_3, dchi_3S, dchi_3NS, dchi_4, dchi_4S, dchi_4NS,
        dchi_5l, dchi_5lS, dchi_5lNS, dchi_6, dchi_6S, dchi_6NS, dchi_6l, dchi_7, dchi_7S, dchi_7NS,
        dchi_kappaS, dchi_kappaA, f_window_div_f_peak, NumCycles

    This involves popping a lot of things from parameters.
    The keys in added_keys should be popped after evaluating the waveform.

    Parameters
    ==========
    parameters: dict
        dictionary of parameter values to convert into the required parameters

    Returns
    =======
    converted_parameters: dict
        dict of the required parameters
    added_keys: list
        keys which are added to parameters during function call
    """

    converted_parameters, added_keys = conversion.convert_to_lal_binary_black_hole_parameters(parameters)
    converted_parameters, added_keys = generate_fti_parameters(converted_parameters, added_keys)

    FTI_params = ['dchi_Minus2', 'dchi_0', 'dchi_1', 'dchi_2', 'dchi_3', 'dchi_3S', 'dchi_3NS',
                  'dchi_4', 'dchi_4S', 'dchi_4NS', 'dchi_5l', 'dchi_5lS', 'dchi_5lNS', 'dchi_6',
                  'dchi_6S', 'dchi_6NS', 'dchi_6l', 'dchi_7', 'dchi_7S', 'dchi_7NS',
                  'dchi_kappaS', 'dchi_kappaA', 'dchi_kappa1', 'dchi_kappa2']
    for key in FTI_params:
        if key not in converted_parameters.keys():
            converted_parameters[key] = 0
            added_keys = added_keys + [key]

    return converted_parameters, added_keys


def generate_all_bbh_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    (Adapted from bilby)
    From either a single sample or a set of samples fill in all missing BBH parameters

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    """

    output_sample = conversion.generate_all_bbh_parameters(sample, likelihood, priors, npool)
    output_sample, _ = generate_fti_parameters(output_sample)

    return output_sample


def convert_to_lal_binary_neutron_star_parameters(parameters):
    """
    (Adapted from bilby)
    Convert parameters we have into parameters we need.

    This is defined by the parameters of bilby_tgr.fti.source.lal_binary_neutron_star()

    Mass: mass_1, mass_2
    Spin: a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl
    Extrinsic: luminosity_distance, theta_jn, phase, ra, dec, geocent_time, psi
    Tidal: lambda_1, lamda_2
    FTI: dchi_Minus2, dchi_0, dchi_1, dchi_2, dchi_3, dchi_3S, dchi_3NS, dchi_4, dchi_4S, dchi_4NS,
        dchi_5l, dchi_5lS, dchi_5lNS, dchi_6, dchi_6S, dchi_6NS, dchi_6l, dchi_7, dchi_7S, dchi_7NS,
        dchi_kappaS, dchi_kappaA, f_window_div_f_peak, NumCycles

    This involves popping a lot of things from parameters.
    The keys in added_keys should be popped after evaluating the waveform.

    Parameters
    ==========
    parameters: dict
        dictionary of parameter values to convert into the required parameters

    Returns
    =======
    converted_parameters: dict
        dict of the required parameters
    added_keys: list
        keys which are added to parameters during function call
    """

    converted_parameters, added_keys = conversion.convert_to_lal_binary_neutron_star_parameters(parameters)
    converted_parameters, added_keys = generate_fti_parameters(converted_parameters, added_keys)

    FTI_params = ['dchi_Minus2', 'dchi_0', 'dchi_1', 'dchi_2', 'dchi_3', 'dchi_3S', 'dchi_3NS',
                  'dchi_4', 'dchi_4S', 'dchi_4NS', 'dchi_5l', 'dchi_5lS', 'dchi_5lNS', 'dchi_6',
                  'dchi_6S', 'dchi_6NS', 'dchi_6l', 'dchi_7', 'dchi_7S', 'dchi_7NS',
                  'dchi_kappaS', 'dchi_kappaA', 'dchi_kappa1', 'dchi_kappa2']
    for key in FTI_params:
        if key not in converted_parameters.keys():
            converted_parameters[key] = 0
            added_keys = added_keys + [key]

    return converted_parameters, added_keys


def generate_all_bns_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    (Adapted from bilby)
    From either a single sample or a set of samples fill in all missing BBH parameters

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    """

    output_sample = conversion.generate_all_bns_parameters(sample, likelihood, priors, npool)
    output_sample, _ = generate_fti_parameters(output_sample)

    return output_sample



import bilby
import numpy as np
from bilby.core.utils.constants import speed_of_light as c


def ssb_factor(
        frequency_array, ssb_coeff, **kwargs):
    """
    Compute the modified dispersion of the waveform due to spacetime symmetry breaking (SSB)
    as described in Eq.(12) of arxiv:2210.04481 (only mass-dimension 5 terms in the Lagrangian density are considered)

    Parameters:

    frequency_array: array of floats
        frequencies on which to evaluate the correction [unit: s^{-1}]
    ssb_coeff: float
        modified dispersion coefficient [unit: m*s]

    Output

    correction: np.array
        SSB correction factor beta to apply to cross and plus polarizations
    """

    beta = 2 * np.pi**(1.5) * ssb_coeff / c
    return beta * frequency_array**2


def lal_binary_black_hole(
        frequency_array, mass_1, mass_2, luminosity_distance, a_1, tilt_1,
        phi_12, a_2, tilt_2, phi_jl, theta_jn, phase, ssb_coeff, **kwargs):
    """
    Binary black hole waveform with modified dispersion due to spacetime symmetry breaking (SSB)
    as described in arxiv:2210.04481. Modified dispersion corresponding to only mass-dimension 5
    terms in the Lagrangian density are considered.

    SSB extra parameters:

    ssb_coeff: float
        modified dispersion coefficient [unit: m*s]
    """

    gr_waveform = bilby.gw.source.lal_binary_black_hole(
        frequency_array=frequency_array, mass_1=mass_1, mass_2=mass_2, luminosity_distance=luminosity_distance,
        a_1=a_1, tilt_1=tilt_1, phi_12=phi_12, a_2=a_2, tilt_2=tilt_2, phi_jl=phi_jl,
        theta_jn=theta_jn, phase=phase, **kwargs)

    ssb_correction = ssb_factor(frequency_array, ssb_coeff, **kwargs)

    if gr_waveform is None:
        return None
    else:
        return dict(plus=gr_waveform['plus'] * np.cos(ssb_correction) - gr_waveform['cross'] * np.sin(ssb_correction),
                    cross=gr_waveform['cross'] * np.cos(ssb_correction) + gr_waveform['plus'] * np.sin(ssb_correction))



import bilby.gw.prior as prior
import os
from bilby.core.utils import logger

DEFAULT_PRIOR_DIR = os.path.join(os.path.dirname(__file__), 'prior_files')


class BBHPriorDict(prior.BBHPriorDict):
    def __init__(self, dictionary=None, filename=None, aligned_spin=False,
                 conversion_function=None, v4_version=False):
        """ Initialises a Prior set for Binary Black holes with pSEOB parameter.
        Adapted from bilby.

        Parameters
        ==========
        dictionary: dict, optional
            See superclass
        filename: str, optional
            See superclass
        conversion_function: func
            Function to convert between sampled parameters and constraints.
            By default this generates many additional parameters, see
            BBHPriorDict.default_conversion_function
        """
        if dictionary is None and filename is None:
            try:
                prior_files = {
                    (True, False): ('aligned_spins_bbh.prior', 'Using aligned-spin prior for pSEOBNRv5HM'),
                    (True, True): ('aligned_spins_bbh_v4.prior', 'Using aligned-spin prior for pSEOBNRv4HM_PA'),
                    (False, False): ('precessing_spins_bbh.prior', 'Using precessing-spin prior for pSEOBNRv5PHM'),
                }
                fname, log_message = prior_files[(aligned_spin, v4_version)]
                logger.info(log_message)

            except KeyError:
                raise NotImplementedError("A precessing-spin prior is not supported for pSEOBNRv4HM_PA")

            filename = os.path.join(DEFAULT_PRIOR_DIR, fname)
            logger.info(f'No prior given, using default BBH priors in {filename}')
        elif filename is not None:
            if not os.path.isfile(filename):
                filename = os.path.join(DEFAULT_PRIOR_DIR, filename)
        super(prior.BBHPriorDict, self).__init__(dictionary=dictionary, filename=filename,
                                                 conversion_function=conversion_function)


# (Adapted from bilby)
prior.Prior._default_latex_labels = {
    'mass_1': '$m_1$',
    'mass_2': '$m_2$',
    'total_mass': '$M$',
    'chirp_mass': r'$\mathcal{M}$',
    'mass_ratio': '$q$',
    'symmetric_mass_ratio': r'$\eta$',
    'a_1': '$a_1$',
    'a_2': '$a_2$',
    'tilt_1': r'$\theta_1$',
    'tilt_2': r'$\theta_2$',
    'cos_tilt_1': r'$\cos\theta_1$',
    'cos_tilt_2': r'$\cos\theta_2$',
    'phi_12': r'$\Delta\phi$',
    'phi_jl': r'$\phi_{JL}$',
    'luminosity_distance': '$d_L$',
    'dec': r'$\mathrm{DEC}$',
    'ra': r'$\mathrm{RA}$',
    'iota': r'$\iota$',
    'cos_iota': r'$\cos\iota$',
    'theta_jn': r'$\theta_{JN}$',
    'cos_theta_jn': r'$\cos\theta_{JN}$',
    'psi': r'$\psi$',
    'phase': r'$\phi$',
    'geocent_time': '$t_c$',
    'time_jitter': '$t_j$',
    'lambda_1': r'$\Lambda_1$',
    'lambda_2': r'$\Lambda_2$',
    'lambda_tilde': r'$\tilde{\Lambda}$',
    'delta_lambda_tilde': r'$\delta\tilde{\Lambda}$',
    'chi_1': r'$\chi_1$',
    'chi_2': r'$\chi_2$',
    'chi_1_in_plane': r'$\chi_{1, \perp}$',
    'chi_2_in_plane': r'$\chi_{2, \perp}$',
    'dchi_Minus2': r'$\delta\varphi_{-2}$',
    'domega220': r'$\delta f_{220}$',
    'domega210': r'$\delta f_{210}$',
    'domega330': r'$\delta f_{330}$',
    'domega320': r'$\delta f_{320}$',
    'domega440': r'$\delta f_{440}$',
    'domega430': r'$\delta f_{430}$',
    'domega550': r'$\delta f_{550}$',
    'dtau220': r'$\delta \tau_{220}$',
    'dtau210': r'$\delta \tau_{210}$',
    'dtau330': r'$\delta \tau_{330}$',
    'dtau320': r'$\delta \tau_{320}$',
    'dtau440': r'$\delta \tau_{440}$',
    'dtau430': r'$\delta \tau_{430}$',
    'dtau550': r'$\delta \tau_{550}$',
    'dA22': r'$\delta A_{22}$',
    'dA21': r'$\delta A_{21}$',
    'dA33': r'$\delta A_{33}$',
    'dA32': r'$\delta A_{32}$',
    'dA44': r'$\delta A_{44}$',
    'dA43': r'$\delta A_{43}$',
    'dA55': r'$\delta A_{55}$',
    'dw22': r'$\delta \omega_{22}$',
    'dw21': r'$\delta \omega_{21}$',
    'dw33': r'$\delta \omega_{33}$',
    'dw32': r'$\delta \omega_{32}$',
    'dw44': r'$\delta \omega_{44}$',
    'dw43': r'$\delta \omega_{43}$',
    'dw55': r'$\delta \omega_{55}$',
    'da6': r'$\delta a_{6}$',
    'ddSO': r'$\delta d_{\mathrm{SO}}$',
    'dTpeak': r'$\delta \Delta_t_{\mathrm{peak}}$',
}



import astropy.units as u
from astropy.cosmology import z_at_value
from bilby.gw.cosmology import get_cosmology
import bilby
from scipy import integrate
import numpy as np


def isiterable(obj):
    """
    Returns True if the given object is iterable.
    Copied from astropy, as location varies between versions
    """
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def vectorize_if_needed(func, *x):
    """
    Helper function to vectorize functions on array inputs
    Copied here from astropy, because vectorization differs between astropy versions
    """
    if any(map(isiterable, x)):
        return np.vectorize(func)(*x)
    else:
        return func(*x)


def luminosity_distance_to_redshift_scalar(distance, cosmology=None):
    """
    copied from bilby, but non-vectorized
    """
    cosmology = get_cosmology(cosmology)

    return z_at_value(cosmology.luminosity_distance, distance * u.Mpc).value


def kv5_from_ssb_coeff_scalar(ssb_coeff, z, cosmology=None):
    """
    converts modified dispersion coefficient to the SME parameter kv5_00 as
    in Eq. 13 of https://arxiv.org/pdf/2210.04481.pdf

    Parameters:

    ssb_coeff: float
        modified dispersion coefficient [unit: m*s]
    z: float
        redshift
    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
        None - Use DEFAULT_COSMOLOGY from bilby
        Instance of astropy.cosmology.FLRW subclass
        String with name of known Astropy cosmology, e.g., Planck13
        Dictionary with arguments required to instantiate the cosmology
        class.

    Returns:

    kv5:  float
        SME parameter [unit: m]
    """
    cosmo = get_cosmology(cosmology)

    # compute tau
    def tau_fct(z):
        return (1 + z) / (cosmo.H(z) / u.Mpc.to(u.km) * u.megaparsec).value
    tau = integrate.quad(tau_fct, 0, z)

    # convert effective parameter
    kv5 = ssb_coeff / tau[0]

    return kv5


def ssb_coeff_from_kv5_scalar(kv5, z, cosmology=None):
    """
    The inverse of ssb_coeff_from_kv5_scalar

    Parameters:

    kv5:  float
        SME parameter [unit: m]
    z: float
        redshift
    cosmology: astropy.cosmology.FLRW, str, dict
        Description of cosmology, one of:
        None - Use DEFAULT_COSMOLOGY from bilby
        Instance of astropy.cosmology.FLRW subclass
        String with name of known Astropy cosmology, e.g., Planck13
        Dictionary with arguments required to instantiate the cosmology
        class.

    Returns:

    ssb_coeff: float
        modified dispersion coefficient [unit: m*s]
    """
    cosmo = get_cosmology(cosmology)

    # compute tau
    def tau_fct(z):
        return (1 + z) / (cosmo.H(z) / u.Mpc.to(u.km) * u.megaparsec).value
    tau = integrate.quad(tau_fct, 0, z)

    # convert effective parameter
    ssb_coeff = kv5 * tau[0]

    return ssb_coeff


def luminosity_distance_to_redshift(distance, cosmology=None):
    """
    copied from bilby, but non-vectorized
    """
    if isiterable(distance):
        distance = np.asarray(distance)
    return vectorize_if_needed(luminosity_distance_to_redshift_scalar, distance, cosmology)


def kv5_from_ssb_coeff(ssb_coeff, z, cosmology=None):
    """
    vectorized version of kv5_from_ssb_coeff_scalar
    """
    if isiterable(z):
        z = np.asarray(z)
    if isiterable(ssb_coeff):
        ssb_coeff = np.asarray(ssb_coeff)
    return vectorize_if_needed(kv5_from_ssb_coeff_scalar, ssb_coeff, z, cosmology)


def ssb_coeff_from_kv5(kv5, z, cosmology=None):
    """
    vectorized version of ssb_coeff_from_kv5_scalar
    """
    if isiterable(z):
        z = np.asarray(z)
    if isiterable(kv5):
        kv5 = np.asarray(kv5)
    return vectorize_if_needed(ssb_coeff_from_kv5_scalar, kv5, z, cosmology)


def generate_ssb_parameters(sample, cosmology=None):

    output_sample = sample.copy()
    redshift = None
    if 'redshift' in output_sample:
        redshift = output_sample['redshift']
    if redshift is None:
        redshift = bilby.gw.conversion.luminosity_distance_to_redshift(output_sample['luminosity_distance'], cosmology)
    if 'ssb_coeff' in output_sample:
        output_sample['kv5'] = kv5_from_ssb_coeff(output_sample['ssb_coeff'], redshift, cosmology=cosmology)
    else :
        print('could not generate samples of kv5')

    return output_sample


def generate_all_bbh_parameters(sample, likelihood=None, priors=None, npool=1):
    """
    adapted from bilby
    From either a single sample or a set of samples fill in all missing
    BBH parameters, in place.

    Parameters
    ==========
    sample: dict or pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    """
    output_sample = bilby.gw.conversion.generate_all_bbh_parameters(sample, likelihood, priors, npool)
    output_sample = generate_ssb_parameters(output_sample)
    return output_sample



import unittest
import glob

import bilby_tgr


class TestPackagedPriors(unittest.TestCase):
    """ Test that the prepackaged priors load """

    def test_aligned(self):
        filename = "aligned_spins_bbh.prior"
        prior_dict = bilby_tgr.pseob.prior.BBHPriorDict(filename=filename)
        self.assertTrue("chi_1" in prior_dict)
        self.assertTrue("chi_2" in prior_dict)
        self.assertTrue("dA22" in prior_dict)

    def test_aligned_v4(self):
        filename = "aligned_spins_bbh_v4.prior"
        prior_dict = bilby_tgr.pseob.prior.BBHPriorDict(filename=filename)
        self.assertTrue("chi_1" in prior_dict)
        self.assertTrue("chi_2" in prior_dict)
        self.assertFalse("dA22" in prior_dict)

    def test_binary_black_holes(self):
        filename = "precessing_spins_bbh.prior"
        prior_dict = bilby_tgr.pseob.prior.BBHPriorDict(filename=filename)
        self.assertTrue("a_1" in prior_dict)
        self.assertTrue("dA22" in prior_dict)

    def test_all(self):
        prior_files = glob.glob(bilby_tgr.pseob.prior.DEFAULT_PRIOR_DIR + "/*prior")
        for ff in prior_files:
            print("Checking prior file {}".format(ff))
            prior_dict = bilby_tgr.pseob.prior.BBHPriorDict(filename=ff)
            self.assertTrue("chirp_mass" in prior_dict)
            self.assertTrue("mass_ratio" in prior_dict)
            if "precessing" in ff:
                self.assertTrue("a_1" in prior_dict)
            elif "aligned" in ff:
                self.assertTrue("chi_1" in prior_dict)


if __name__ == "__main__":
    unittest.main()



from __future__ import division, print_function

import bilby
import numpy as np

import bilby_tgr

outdir = "outdir"
label = "siqm_example"
bilby.core.utils.setup_logger(outdir=outdir, label=label)
np.random.seed(88170235)
injection_parameters = dict(
    mass_1=20.0,
    mass_2=10.0,
    a_1=0.7,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=200.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
    dQuadMon1=20,
    dQuadMon2=0,
)
duration = 4
sampling_frequency = 2 * 1024
start_time = injection_parameters["geocent_time"] + 2 - duration
waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2",
    reference_frequency=50.0,
    minimum_frequency=40.0,
)


waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby_tgr.siqm.source.lal_binary_black_hole,
    parameter_conversion=bilby_tgr.siqm.conversion.convert_to_lal_siqm_parameters,
    waveform_arguments=waveform_arguments,
)

# (LIGO-Hanford (H1), LIGO-Livingston (L1), and Virgo (V1)).
# These default to their design sensitivity and start at 40 Hz.

ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 3,
)
ifos.inject_signal(
    waveform_generator=waveform_generator, parameters=injection_parameters
)

priors = bilby.gw.prior.BBHPriorDict()
priors["geocent_time"] = bilby.core.prior.Uniform(
    minimum=injection_parameters["geocent_time"] - 1,
    maximum=injection_parameters["geocent_time"] + 1,
    name="geocent_time",
    latex_label="$t_c$",
    unit="$s$",
)
for key in [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "phi_jl",
    "psi",
    "ra",
    "dec",
    "geocent_time",
    "phase",
]:
    priors[key] = injection_parameters[key]
priors["dQuadMonS"] = bilby.core.prior.Uniform(0, 500, name="dQuadMonS")
priors["dQuadMonA"] = 0
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)

result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    npoints=1000,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# Make a corner plot.
result.plot_corner()
